import json
import unittest

import yaml

from tira.check_format import check_format, lines_if_valid

from . import (
    _ERROR,
    _OK,
    EMPTY_OUTPUT,
    IR_METADATA_INVALID,
    IR_METADATA_MULTIPLE_VALID,
    IR_METADATA_SINGLE_VALID,
    IR_QUERY_OUTPUT,
    TSV_OUTPUT_VALID,
    VALID_RUN_OUTPUT,
)


class TestCheckTsvForNonExistingFormats(unittest.TestCase):
    def test_invalid_validator_on_empty_output(self):
        expected = [_ERROR, "At least one valid ir_metadata file was expected, but there was none."]
        actual = check_format(EMPTY_OUTPUT, "ir_metadata")
        self.assertEqual(expected, actual)

    def test_invalid_validator_on_query_output(self):
        expected = [_ERROR, "At least one valid ir_metadata file was expected, but there was none."]
        actual = check_format(IR_QUERY_OUTPUT, "ir_metadata")
        self.assertEqual(expected, actual)

    def test_invalid_trec_run_file(self):
        expected = [_ERROR, "At least one valid ir_metadata file was expected, but there was none."]
        actual = check_format(VALID_RUN_OUTPUT, "ir_metadata")
        self.assertEqual(expected, actual)

    def test_invalid_tsv(self):
        expected = [_ERROR, "At least one valid ir_metadata file was expected, but there was none."]
        actual = check_format(TSV_OUTPUT_VALID, "ir_metadata")
        self.assertEqual(expected, actual)

    def test_invalid_ir_metadata(self):
        with open(IR_METADATA_INVALID / ".ir-metadata.yml", "r") as f:
            actual = yaml.safe_load(f)
            self.assertEqual({"test": {"foo": "Hello"}}, actual)

        expected = [_ERROR, "At least one valid ir_metadata file was expected, but there was none."]
        actual = check_format(IR_METADATA_INVALID, "ir_metadata")
        self.assertEqual(expected, actual)

    def test_valid_output_with_single_ir_metadata(self):
        expected = [_OK, "The output provides valid ir_metadata."]
        actual = check_format(IR_METADATA_SINGLE_VALID, "ir_metadata")
        self.assertEqual(expected, actual)

    def test_valid_output_with_multiple_ir_metadata(self):
        expected = [_OK, "The output provides valid ir_metadata."]
        actual = check_format(IR_METADATA_MULTIPLE_VALID, "ir_metadata")
        self.assertEqual(expected, actual)

    def test_lines_of_valid_single_ir_metadata(self):
        expected = ".ir-metadata.yml"
        actual = lines_if_valid(IR_METADATA_SINGLE_VALID, "ir_metadata")
        self.assertEqual(1, len(actual))
        self.assertEqual(expected, actual[0]["name"])
        self.assertEqual(json.loads(json.dumps(actual[0]["content"])), actual[0]["content"])

    def test_lines_of_valid_multiple_ir_metadata(self):
        actual = lines_if_valid(IR_METADATA_MULTIPLE_VALID, "ir_metadata")
        self.assertEqual(6, len(actual))
        for i in actual:
            self.assertIsNotNone(i["name"])
            self.assertEqual(json.loads(json.dumps(i["content"])), i["content"])



# from approvaltests import verify_as_json
import os
import sys
import tempfile
import unittest

from tests.jupyter_notebook_pipeline_construction_test import build_docker_image
from tira.tira_run import main


def foo(image):
    orig_argv = list(sys.argv)
    temp_dir = "/tmp/" + tempfile.TemporaryDirectory().name
    os.makedirs(temp_dir, exist_ok=True)

    sys.argv = [
        "tira-run",
        "--input-dataset",
        "workshop-on-open-web-search/document-processing-20231027-training",
        "--image",
        image,
        "--output-dir",
        "fffff",
    ]
    main()
    sys.argv = orig_argv


class TiraRunIntegrationTest(unittest.TestCase):

    def testfoo(self):
        build_docker_image("dockerfile_multi_stage_bash")

        # foo(image)
        # raise ValueError('sda')



import unittest
from pathlib import Path

from tira.local_execution_integration import LocalExecutionIntegration as Client
from tira.third_party_integrations import (
    extract_previous_stages_from_docker_image,
    extract_previous_stages_from_notebook,
    extract_to_be_executed_notebook_from_command_or_none,
    parse_ast_extract_assignment,
    parse_extraction_of_tira_approach,
    parse_extraction_of_tira_approach_bash,
)

TEST_DIR = Path(__file__).parent.resolve()


def build_docker_image(image):
    Client()._LocalExecutionIntegration__docker_client().images.build(
        path=".", dockerfile="tests/resources/" + image, tag=image
    )


class JupyterNotebookPipelineConstructionTest(unittest.TestCase):
    def test_no_notebook_is_extracted_for_none_command(self):
        command = None
        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertIsNone(actual)

    def test_no_notebook_is_extracted_for_empty_command(self):
        command = ""
        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertIsNone(actual)

    def test_notebook_is_extracted_for_pyterrier_command(self):
        expected = "/workspace/notebook.ipynb"
        command = (
            "/workspace/run-pyterrier-notebook.py --input $inputDataset --output $outputDir --notebook"
            " /workspace/notebook.ipynb"
        )

        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertEqual(expected, actual)

    def test_notebook_is_extracted_when_subshell_is_available(self):
        expected = "/re-rank-with-tiny-bert.ipynb"
        command = (
            'bash -c export MODEL="$(huggingface-cli download $MODEL)" && /run-notebook.py --notebook'
            " /re-rank-with-tiny-bert.ipynb"
        )

        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertEqual(expected, actual)

    def test_notebook_is_extracted_for_command_in_between(self):
        expected = "/notebook.ipynb"
        command = "/workspace/run-notebook.py --notebook /notebook.ipynb --input $inputDataset --output $outputDir"

        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertEqual(expected, actual)

    def test_py_file_is_extracted_for_command_in_between(self):
        expected = "app.py"
        command = "python3 app.py"

        actual = extract_to_be_executed_notebook_from_command_or_none(command)

        self.assertEqual(expected, actual)

    def test_no_previous_stages_are_extracted_from_notebook_without_previous_stages(self):
        notebook = TEST_DIR / "resources" / "pyterrier-notebook-without-previous-stages.ipynb"
        expected = []

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_no_previous_stages_are_extracted_from_python_file_without_previous_stages(self):
        notebook = TEST_DIR / "resources" / "pyterrier-notebook-without-previous-stages.py"
        expected = []

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_pyterrier_index_as_previous_stage(self):
        notebook = TEST_DIR / "resources" / "retrieve-with-pyterrier-index.ipynb"
        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_for_query_transformation_01(self):
        python_line = (
            "gpt_sq_zs = tira.pt.transform_queries('ir-benchmarks/tu-dresden-03/qe-gpt3.5-sq-zs', dataset,"
            " prefix='llm_expansion_')"
        )
        expected = "ir-benchmarks/tu-dresden-03/qe-gpt3.5-sq-zs"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_for_document_transformation_01(self):
        python_line = (
            "document_entity_recognition = tira.pt.transform_documents('ir-lab-sose-2024/ir-nfmj/entity-recognition',"
            " pt_dataset)"
        )
        expected = "ir-lab-sose-2024/ir-nfmj/entity-recognition"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_llm_expansions_as_previous_stage(self):
        notebook = TEST_DIR / "resources" / "jupyter-notebook-with-query-expansion.ipynb"
        expected = [
            "ir-lab-sose-2024/tira-ir-starter/Index (tira-ir-starter-pyterrier)",
            "ir-benchmarks/tu-dresden-03/qe-gpt3.5-sq-zs",
        ]

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_pyterrier_index_as_previous_stage_for_python_file(self):
        notebook = TEST_DIR / "resources" / "retrieve-with-pyterrier-index.py"
        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_pyterrier_index_as_previous_stage_for_python_file_with_utility_function(self):
        python_line = (
            '    index = tira.pt.index("ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)", pt_dataset)'
        )
        expected = "ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_pyterrier_index_as_previous_stage_for_python_file_with_utility_function_complete(self):
        notebook = TEST_DIR / "resources" / "corpus-graph-with-pyterrier-index.py"
        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_parsing_of_ast_assignment_none(self):
        python_line = None
        k, v = parse_ast_extract_assignment(python_line)

        self.assertIsNone(k)
        self.assertIsNone(v)

    def test_parsing_of_ast_assignment_empty_string(self):
        python_line = None
        k, v = parse_ast_extract_assignment(python_line)

        self.assertIsNone(k)
        self.assertIsNone(v)

    def test_parsing_of_ast_assignment_method_call(self):
        python_line = (
            "index = tira.pt.index('ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)',"
            " 'longeval-tiny-train-20240315-training')"
        )
        k, v = parse_ast_extract_assignment(python_line)

        self.assertIsNone(k)
        self.assertIsNone(v)

    def test_parsing_of_ast_assignment_to_constant(self):
        python_line = "index = 3"
        expected_k, expeced_v = "index", 3
        k, v = parse_ast_extract_assignment(python_line)

        self.assertEqual(expected_k, k)
        self.assertEqual(expeced_v, v)

    def test_extraction_of_approach_is_failsave_01(self):
        python_line = "index = None"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_is_failsave_02(self):
        python_line = ""
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_is_failsave_03(self):
        python_line = None
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_is_failsave_04(self):
        python_line = "#index = tira.pt.index('tmp', 'longeval-tiny-train-20240315-training') # some comment"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_01(self):
        python_line = (
            "index = tira.pt.index('ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)',"
            " 'longeval-tiny-train-20240315-training')"
        )
        expected = "ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_02(self):
        python_line = "index = tira.pt.index('tmp', 'longeval-tiny-train-20240315-training') # some comment"
        expected = "tmp"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_03(self):
        python_line = 'index = tira.pt.index("tmp\'\\"a","longeval-tiny-train-20240315-training") # some comment'
        expected = "tmp'\"a"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_integration_against_docker_image_01(self):
        image = "mam10eks/bash-with-notebooks:latest"
        command = (
            "bash /workspace/run-notebook.sh --notebook /pyterrier-notebook-without-previous-stages.ipynb --input"
            " $inputDataset --output $outputDir"
        )

        expected = []
        actual = extract_previous_stages_from_docker_image(image, command)

        self.assertEqual(expected, actual)

    def test_integration_against_docker_image_02(self):
        image = "mam10eks/bash-with-notebooks:latest"
        command = (
            "bash /workspace/run-notebook.sh  --input $inputDataset --output $outputDir --notebook"
            " /pyterrier-notebook-without-previous-stages.ipynb"
        )

        expected = []
        actual = extract_previous_stages_from_docker_image(image, command)

        self.assertEqual(expected, actual)

    def test_integration_against_docker_image_03(self):
        image = "mam10eks/bash-with-notebooks:latest"
        command = (
            "bash /workspace/run-notebook.sh  --input $inputDataset --output $outputDir --notebook"
            " /retrieve-with-pyterrier-index.ipynb"
        )

        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]
        actual = extract_previous_stages_from_docker_image(image, command)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_non_valid_input_01(self):
        bash_line = ""
        actual = parse_extraction_of_tira_approach_bash(bash_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_from_bash_line_non_valid_input_02(self):
        bash_line = None
        actual = parse_extraction_of_tira_approach_bash(bash_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_from_bash_line_non_valid_input_03(self):
        bash_line = "tira-cli --help"
        actual = parse_extraction_of_tira_approach_bash(bash_line)

        self.assertIsNone(actual)

    def test_extraction_of_approach_from_bash_line_01(self):
        bash_line = (
            "INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach tmp) # some comment"
        )
        expected = "tmp"
        actual = parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_02(self):
        bash_line = (
            "INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach"
            " 'ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)') # some comment"
        )
        expected = "ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"
        actual = parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_03(self):
        bash_line = (
            'INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach "tmp") # some comment'
        )
        expected = "tmp"
        actual = parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_04(self):
        bash_line = (
            "INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach"
            ' "ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)") # some comment'
        )
        expected = "ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"
        actual = parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_extraction_of_approach_from_bash_line_05(self):
        bash_line = (
            "INDEX=$(tira-cli download --dataset longeval-tiny-train-20240315-training --approach 'tmp') # some comment"
        )
        expected = "tmp"
        actual = parse_extraction_of_tira_approach_bash(bash_line)

        self.assertEqual(expected, actual)

    def test_pyterrier_index_as_previous_stage_for_bash_file(self):
        notebook = TEST_DIR / "resources" / "retrieve-with-pyterrier-bash.sh"
        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_no_previous_stage_for_bash_file(self):
        notebook = TEST_DIR / "resources" / "retrieve-no-previous-stages-bash.sh"
        expected = []

        actual = extract_previous_stages_from_notebook(notebook)

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_01(self):
        image = "dockerfile_bash_script_absolute"
        build_docker_image(image)

        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]
        actual = extract_previous_stages_from_docker_image(image, "/usr/bin/retrieve-with-pyterrier-bash.sh")

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_02(self):
        image = "dockerfile_bash_script_absolute"
        build_docker_image(image)

        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]
        actual = extract_previous_stages_from_docker_image(image)

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_03(self):
        image = "dockerfile_bash_script_absolute"
        build_docker_image(image)

        expected = []
        actual = extract_previous_stages_from_docker_image(image, "/etc/hostname")

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_04(self):
        image = "jupyter_script_relative"
        build_docker_image(image)

        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]
        actual = extract_previous_stages_from_docker_image(image)

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_05(self):
        image = "jupyter_script_relative"
        build_docker_image(image)

        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]
        actual = extract_previous_stages_from_docker_image(image, "python3 /usr/bin/retrieve-with-pyterrier-index.py")

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_06(self):
        image = "jupyter_script_relative"
        build_docker_image(image)

        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]
        actual = extract_previous_stages_from_docker_image(
            image, "python3 /usr/bin/retrieve-with-pyterrier-index.py; sleep3"
        )

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_07(self):
        image = "jupyter_script_relative"
        build_docker_image(image)

        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]
        actual = extract_previous_stages_from_docker_image(
            image, "python3 /usr/bin/retrieve-with-pyterrier-index.py&&sleep3"
        )

        self.assertEqual(expected, actual)

    def test_integration_against_custom_docker_image_08(self):
        image = "jupyter_script_relative"
        build_docker_image(image)

        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]
        actual = extract_previous_stages_from_docker_image(
            image, 'python3 "/usr/bin/retrieve-with-pyterrier-index.py"&&sleep3'
        )

        self.assertEqual(expected, actual)
        self.assertEqual(
            '"/usr/bin/retrieve-with-pyterrier-index.py"',
            extract_to_be_executed_notebook_from_command_or_none(
                'python3 "/usr/bin/retrieve-with-pyterrier-index.py"&&sleep3'
            ),
        )

    def test_integration_against_custom_docker_image_09(self):
        image = "jupyter_script_relative"
        build_docker_image(image)

        expected = ["ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)"]
        actual = extract_previous_stages_from_docker_image(
            image, "python3 '/usr/bin/retrieve-with-pyterrier-index.py'&&sleep3"
        )

        self.assertEqual(expected, actual)
        self.assertEqual(
            "'/usr/bin/retrieve-with-pyterrier-index.py'",
            extract_to_be_executed_notebook_from_command_or_none(
                "python3 '/usr/bin/retrieve-with-pyterrier-index.py'&&sleep3"
            ),
        )

    def test_pyterrier_from_submission_with_percent_operator_01(self):
        python_line = (
            "        bm25_raw = tira.pt.from_submission('ir-benchmarks/tira-ir-starter/BM25"
            " (tira-ir-starter-pyterrier)', args.input_dataset) % args.first_stage_top_k"
        )
        expected = "ir-benchmarks/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_pyterrier_from_submission_with_percent_operator_02(self):
        notebook = TEST_DIR / "resources" / "keyquery-script.py"
        expected = []

        actual = extract_previous_stages_from_notebook(notebook)
        expected = [
            "ir-benchmarks/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)",
            "ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)",
            "ir-benchmarks/tira-ir-starter/Index (tira-ir-starter-pyterrier)",
        ]

        self.assertEqual(expected, actual)

    def test_pyterrier_from_run_output_01(self):
        python_line = (
            "        bm25_raw = tira.get_run_output('ir-benchmarks/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)',"
            " args.input_dataset)"
        )
        expected = "ir-benchmarks/tira-ir-starter/BM25 (tira-ir-starter-pyterrier)"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)

    def test_pyterrier_from_run_output_02(self):
        python_line = (
            "        plaid_index = tira.get_run_output('reneuir-2024/reneuir-baselines/plaid-x', dataset_id) + '/index'"
        )
        expected = "reneuir-2024/reneuir-baselines/plaid-x"
        actual = parse_extraction_of_tira_approach(python_line)

        self.assertEqual(expected, actual)



import unittest

from tira.check_format import check_format, lines_if_valid

from . import (
    _ERROR,
    _OK,
    EMPTY_OUTPUT,
    IR_QUERY_OUTPUT,
    JSONL_GZ_OUTPUT_VALID,
    JSONL_OUTPUT_INVALID,
    JSONL_OUTPUT_VALID,
    TSV_OUTPUT_VALID,
    VALID_RUN_OUTPUT,
)


class TestCheckJsonlForNonExistingFormats(unittest.TestCase):
    def test_invalid_validator_on_empty_output(self):
        expected = [_ERROR, "No unique *.jsonl file was found, only the files ['.gitkeep'] were available."]
        actual = check_format(EMPTY_OUTPUT, "*.jsonl")
        self.assertEqual(expected, actual)

    def test_valid_jsonl_output_directory(self):
        expected = [_OK, "The jsonl file has the correct format."]
        actual = check_format(JSONL_OUTPUT_VALID, "*.jsonl")
        self.assertEqual(expected, actual)

    def test_valid_jsonl_output_file(self):
        expected = [_OK, "The jsonl file has the correct format."]
        actual = check_format(JSONL_OUTPUT_VALID / "predictions.jsonl", "*.jsonl")
        self.assertEqual(expected, actual)

    def test_valid_jsonl_gz_output_directory(self):
        expected = [_OK, "The jsonl file has the correct format."]
        actual = check_format(JSONL_GZ_OUTPUT_VALID, "*.jsonl")
        self.assertEqual(expected, actual)

    def test_valid_jsonl_gz_output_file(self):
        expected = [_OK, "The jsonl file has the correct format."]
        actual = check_format(JSONL_GZ_OUTPUT_VALID / "predictions.jsonl.gz", "*.jsonl")
        self.assertEqual(expected, actual)

    def test_invalid_jsonl_output_directory(self):
        actual = check_format(JSONL_OUTPUT_INVALID, "*.jsonl")
        self.assertEqual(actual[0], _ERROR)
        self.assertTrue("contains a line that could not be parsed" in actual[1])

    def test_invalid_jsonl_output_file(self):
        actual = check_format(JSONL_OUTPUT_INVALID / "predictions.jsonl", "*.jsonl")
        self.assertEqual(actual[0], _ERROR)
        self.assertTrue("contains a line that could not be parsed" in actual[1])

    def test_invalid_on_query_output_directory(self):
        actual = check_format(IR_QUERY_OUTPUT, "*.jsonl")
        self.assertEqual(actual[0], _ERROR)
        self.assertTrue('The json line misses the required field "id"' in actual[1])

    def test_invalid_on_query_output_file(self):
        actual = check_format(IR_QUERY_OUTPUT / "queries.jsonl", "*.jsonl")
        self.assertEqual(actual[0], _ERROR)
        self.assertTrue('The json line misses the required field "id"' in actual[1])

    def test_invalid_trec_run_file(self):
        expected = [_ERROR, "No unique *.jsonl file was found, only the files ['run.txt'] were available."]
        actual = check_format(VALID_RUN_OUTPUT, "*.jsonl")
        self.assertEqual(expected, actual)

    def test_invalid_tsv(self):
        expected = [_ERROR, "No unique *.jsonl file was found, only the files ['predictions.tsv'] were available."]
        actual = check_format(TSV_OUTPUT_VALID, "*.jsonl")
        self.assertEqual(expected, actual)



from hashlib import md5

import requests

from tira.rest_api_client import Client
from tira.tira_redirects import redirects


def md5_of_first_kilobyte_of_http_resource(url):
    # if not url.startswith('https://files.webis.de'):
    #    raise ValueError(f'URL {url} is not from webis.de respectively zenodo')
    print(url, flush=True)
    return md5(requests.get(url, headers={"Range": "bytes=0-1024"}).content).hexdigest()


def digest_of_resource(resource):
    from tira.tira_redirects import RESOURCE_REDIRECTS

    if resource not in RESOURCE_REDIRECTS:
        raise ValueError("Could not find resource {resource}.")

    url = RESOURCE_REDIRECTS[resource]

    return {
        "url": url,
        "redirect_url": redirects(url=url),
        "md5_of_first_kilobyte": md5_of_first_kilobyte_of_http_resource(url),
    }


def digest_of_dataset(dataset_id, truth=False):
    t = "input-" if not truth else "input-truth"
    tira_url = f"https://www.tira.io/data-download/training/{t}/{dataset_id}.zip"
    url_redirects = redirects(url=tira_url)

    return {
        "tira_url": tira_url,
        "redirect_url": url_redirects,
        "md5_of_first_kilobyte": md5_of_first_kilobyte_of_http_resource(url_redirects["urls"][0]),
    }


def digest_of_run_output(approach, dataset_id, run_ids):
    tira = Client()
    task, team, system = approach.split("/")
    filename = f"{run_ids[task][team][system][dataset_id]}.zip"
    run_url = f"https://www.tira.io/task/{task}/user/{team}/dataset/{dataset_id}/download/{filename}"

    return {
        "redirect_url": redirects(approach, dataset_id),
        "run_execution": tira.get_run_execution_or_none(approach, dataset_id),
        "run_url": run_url,
        "redirected_run_url": redirects(url=run_url),
        "md5_of_first_kilobyte": md5_of_first_kilobyte_of_http_resource(redirects(url=run_url)["urls"][0]),
    }



import json
import os
import shutil
import tempfile
import uuid
import zipfile
from abc import ABC
from pathlib import Path
from typing import TYPE_CHECKING, Union, overload

from tira.check_format import _fmt, check_format, log_message

if TYPE_CHECKING:
    import io
    from typing import Any, Dict, Optional

    import pandas as pd

# .. todo:: this file needs further documentation


class TiraClient(ABC):
    def all_datasets(self) -> "pd.DataFrame":
        pass

    def all_softwares(self) -> "pd.DataFrame":
        pass

    def print_overview_of_all_software(self) -> None:
        pass

    def all_evaluators(self) -> "pd.DataFrame":
        pass

    def all_evaluated_appraoches(self) -> "pd.DataFrame":
        pass

    def docker_software() -> "Any":
        # .. todo:: typehint
        pass

    def run_was_already_executed_on_dataset(self, approach, dataset) -> bool:
        # .. todo:: typehint
        pass

    def get_run_output(self, approach, dataset, allow_without_evaluation=False) -> str:
        # .. todo:: typehint
        pass

    def get_run_execution_or_none(self, approach, dataset, previous_stage_run_id=None) -> "Optional[Dict[str, str]]":
        # .. todo:: typehint
        pass

    def docker_registry(self):
        return "registry.webis.de"

    def modify_task(self, task_id: str, to_rename: "Dict[str, Any]"):
        # .. todo:: typehint
        pass

    @overload
    def download_run(
        self,
        task,
        dataset,
        software,
        team=None,
        previous_stage=None,
        return_metadata: bool = False,
    ) -> "tuple[pd.DataFrame, str]":
        # .. todo:: typehint
        ...

    @overload
    def download_run(
        self,
        task,
        dataset,
        software,
        team=None,
        previous_stage=None,
        return_metadata: bool = False,
    ) -> "pd.DataFrame":
        # .. todo:: typehint
        ...

    def download_run(
        self,
        task,
        dataset,
        software,
        team=None,
        previous_stage=None,
        return_metadata: bool = False,
    ) -> "Union[pd.DataFrame, tuple[pd.DataFrame, str]]":
        # .. todo:: typehint
        pass

    def create_new_upload(self, task_id: str, vm_id: str) -> "Optional[str]":
        """
        Creates a new upload and returns the newly created id. Returns None on failure.
        """
        pass

    def submit_run(self, task_id: str, vm_id: str, dataset_id: str, upload_id: str, run: "io.IOBase") -> bool:
        """
        Submits the runfile located at `run` for the given task and vm for the given upload id. Returns true on success,
        false  otherwise.
        """
        pass

    def _git_repo(self, path: Path):
        import git

        try:
            # TODO: Replace with call to tirex_tracker.
            return git.Repo(path, search_parent_directories=True)
        except git.exc.InvalidGitRepositoryError as e:
            raise ValueError(f"No git repository found at {path}") from e

    def _zip_tracked_files(self, repo: "git.Repo", directory: str):
        """
        Creates a zip archive containing all tracked files in a given Git repository.

        :param repo: The Git repository.
        :param zip_filename: Name of the output zip file.
        """
        # TODO: Replace with call to tirex_tracker.
        tracked_files = [i.path for i in repo.commit().tree.traverse() if i.path.startswith(f"{directory}/")]
        zip_path = Path(tempfile.TemporaryDirectory().name) / "repo.zip"
        zip_path.parent.mkdir(exist_ok=True, parents=True)

        with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zipf:
            for file in tracked_files:
                file_path = os.path.join(repo.working_tree_dir, file)
                zipf.write(file_path, arcname=file)
        return zip_path

    def evaluate(self, directory: Path, dataset_id: str):
        """Evaluate some predictions made for some dataset on your local machine.

        Args:
            directory (Path): The path to the directory that contains the predictions that you want to evaluate.
            dataset_id (str): The ID of the TIRA dataset on which the directory is to be evaluated.
        """
        from tira.io_utils import load_output_of_directory

        all_messages = []

        def print_message(message, level):
            all_messages.append((message, level))
            os.system("cls" if os.name == "nt" else "clear")
            print(f"TIRA Evaluation on '{dataset_id}':")
            for m, l in all_messages:
                log_message(m, l)

        dataset_handle = self.get_dataset(dataset_id)
        if (
            not dataset_handle
            or "mirrors" not in dataset_handle
            or "inputs" not in dataset_handle["mirrors"]
            or "format" not in dataset_handle
            or len(dataset_handle["format"]) == 0
        ):
            raise ValueError("Dataset configuration is invalid: " + str(dataset_handle))

        if (
            "evaluator" not in dataset_handle
            or "image" not in dataset_handle["evaluator"]
            or "command" not in dataset_handle["evaluator"]
        ):
            raise ValueError("Dataset configuration has no evaluator configured: " + str(dataset_handle))

        if not directory or not directory.exists():
            print_message(f"The directory {directory} does not exist.", _fmt.ERROR)
            raise ValueError(f"The directory {directory} does not exist.")

        result, msg = check_format(directory, dataset_handle["format"][0])
        if result != _fmt.OK:
            print_message(msg, result)
            raise ValueError(msg)
        print_message(f"The predictions in {directory} have the expected format.", _fmt.OK)

        print(dataset_handle)
        dataset_path = self.download_dataset(dataset_handle["default_task"], dataset_id, truth_dataset=True)

        print_message(f"The dataset {dataset_id} is available locally.", _fmt.OK)

        if not self.local_execution.docker_is_installed_failsave():
            msg = "Docker is not installed, I can not run the dockerized evaluator..."
            print_message(msg, _fmt.ERROR)
            raise ValueError(msg)

        print_message("Docker is installed. I can run the dockerized evaluator.", _fmt.OK)
        self.local_execution.ensure_image_available_locally(dataset_handle["evaluator"]["image"])

        print_message(f"The evaluator image {dataset_handle['evaluator']['image']} is available locally.", _fmt.OK)

        with tempfile.TemporaryDirectory() as tmp_dir:
            eval_config = {
                "evaluator_id": "1",
                "evaluator_git_runner_image": dataset_handle["evaluator"]["image"],
                "evaluator_git_runner_command": dataset_handle["evaluator"]["command"],
                "truth_directory": str(Path(dataset_path).absolute().resolve()),
            }
            self.local_execution.evaluate(
                Path(tmp_dir).absolute().resolve(),
                directory.absolute().resolve(),
                allow_network=False,
                evaluate=eval_config,
            )

            preds = None
            try:
                preds = json.dumps(load_output_of_directory(Path(tmp_dir), evaluation=True))
            except:
                msg = "The evaluator failed to produce a valid evaluation..."
                print(msg)
                raise ValueError(msg)

            print_message("The evaluator was successfull.", _fmt.OK)
            print("\n\nResult:\n\t" + preds)

    def submit_code(
        self,
        path: Path,
        task_id: str,
        command: "Optional[str]" = None,
        dataset_id: "Optional[str]" = None,
        user_id: "Optional[str]" = None,
        docker_file: "Optional[Path]" = None,
        dry_run: "Optional[bool]" = False,
        allow_network: "Optional[bool]" = False,
    ):
        """Build a tira submission from a git repository.

        Args:
            path (Path): The path to the directory that contains the submission. The path points to the directory used as root for the docker build and must be in a git repository.
            task_id (str): The ID of the TIRA task to which the submissions should be made.
            dataset_id (str, optional): The ID of the TIRA dataset on which the submission is to be tested. If no dataset is passed, the submission will be tested on a small publicly available smoke test dataset for the task.
            user_id (str, optional): The ID of the TIRA team that makes the submission. Is only required if a user has multiple teams.
            docker_file (Path, optional): The Dockerfile to build the submission within the repository. Defaults to None to use path/Dockerfile.
        """

        all_messages = []

        def print_message(message, level):
            all_messages.append((message, level))
            os.system("cls" if os.name == "nt" else "clear")
            print("TIRA Code Submission:")
            for m, l in all_messages:
                log_message(m, l)

        if not dry_run:
            print(f"Ensure that you are registed for task {task_id}.")
            from tira.tira_run import guess_vm_id_of_user

            user_id = guess_vm_id_of_user(task_id, self, user_id)
            assert user_id
            print_message(f"You are registered and will submit as team {user_id} to task {task_id}.", _fmt.OK)

        if dataset_id is None:
            for k, v in self.datasets(task_id).items():
                if (
                    "is_confidential" not in v
                    or v["is_confidential"]
                    or "is_deprecated" not in v
                    or v["is_deprecated"]
                    or "format" not in v
                    or len(v["format"]) == 0
                ):
                    continue
                candidate = self.get_dataset(f"{task_id}/{k}")
                if not candidate or "mirrors" not in candidate or "inputs" not in candidate["mirrors"]:
                    continue
                dataset_id = k

            if dataset_id is None:
                raise ValueError("foo")

        dataset_handle = self.get_dataset(f"{task_id}/{dataset_id}")
        if (
            not dataset_handle
            or "mirrors" not in dataset_handle
            or "inputs" not in dataset_handle["mirrors"]
            or "format" not in dataset_handle
            or len(dataset_handle["format"]) == 0
        ):
            raise ValueError("foo")

        dataset_path = self.download_dataset(task_id, dataset_id)
        print_message(f"The dataset {dataset_id} is available locally.", _fmt.OK)

        repo = self._git_repo(path)

        try:
            remotes = {remote.name: remote.url for remote in repo.remotes}
        except:
            raise ValueError("No remotes found for the git repository.")

        if len(remotes) == 0:
            raise ValueError("No remotes found for the git repository.")

        try:
            commit = repo.commit().hexsha
        except:
            raise ValueError("No commits in the git repository")

        try:
            active_branch = repo.active_branch.name
        except:
            raise ValueError("No branch in the git repository")

        print_message(f"The code is in a git repository {repo.working_tree_dir}.", _fmt.OK)

        if docker_file is None:
            docker_file = path / "Dockerfile"

        docker_file = Path(docker_file)
        if not docker_file.exists():
            raise ValueError(f"No dockerfile {docker_file} exists.")

        directory_in_path = str(Path(path).absolute()).replace(str(Path(repo.working_tree_dir).absolute()) + "/", "")
        submission_name = directory_in_path.replace("/", "-")
        docker_tag = submission_name + "-" + str(uuid.uuid4())[:5]

        if repo.is_dirty(untracked_files=True):
            if not dry_run:
                raise ValueError("The git repository is not clean.")
        else:
            print_message(f"The git repository {repo.working_tree_dir} is clean.", _fmt.OK)
        print("Build Docker image...")
        zipped_code = self._zip_tracked_files(repo, directory_in_path)

        self.local_execution.build_docker_image(path, docker_tag, docker_file)

        print_message(f"The code is embedded into the docker image {docker_tag}.", _fmt.OK)
        print("Test Docker image...")

        if command is None:
            command = self.local_execution.extract_entrypoint(docker_tag)

        with tempfile.TemporaryDirectory() as tmp_dir:
            self.local_execution.run(
                image=docker_tag,
                command=command,
                input_dir=dataset_path,
                output_dir=tmp_dir,
                allow_network=allow_network,
            )
            result, msg = check_format(Path(tmp_dir), dataset_handle["format"][0])
            if result != _fmt.OK:
                print(msg)
                raise ValueError(msg)
            print_message(f"The docker image produced valid outputs on the dataset {dataset_id}.", _fmt.OK)
            shutil.copy(zipped_code, Path(tmp_dir) / "source-code.zip")

            if not dry_run:
                print("Upload Code Submission image...")
                metadata_uuid = self.upload_run_anonymous(tmp_dir, dataset_id)["uuid"]

                print_message(f"The meta data is uploaded to TIRA.", _fmt.OK)

                print("Push Docker image to TIRA...")
                from tira.tira_run import push_image

                pushed_image = push_image(self, docker_tag, task_id, user_id)

                print_message(f"The Docker image is pushed to TIRA.", _fmt.OK)
                print("Configure code submission in TIRA...")
                upload = self.add_docker_software(
                    pushed_image,
                    command,
                    user_id,
                    task_id,
                    None,
                    dict(os.environ),
                    source_code_remotes=remotes,
                    source_code_commit=commit,
                    source_code_active_branch=active_branch,
                    try_run_metadata_uuid=metadata_uuid,
                )
                print_message(f"The code submission is uploaded to TIRA.", _fmt.OK)
                print("\nResult:")
                log_message(
                    f"Your code submission is available in TIRA as {upload['display_name']}.",
                    _fmt.OK,
                )

        return {
            "code": zipped_code,
            "remotes": remotes,
            "commit": commit,
            "active_branch": active_branch,
            "image": docker_tag,
        }

    def __extract_dataset_identifier(self, dataset: any):
        """Extract the dataset identifier from a passed object.

        Args:
            dataset (any): Some representation of dataset.

        Returns:
            Optional[dict]: The dataset identifier if available.
        """
        if hasattr(dataset, "irds_ref"):
            return self.__extract_dataset_identifier(dataset.irds_ref())
        if hasattr(dataset, "dataset_id"):
            return dataset.dataset_id()
        return dataset

    def __matching_dataset(self, datasets, dataset_identifier) -> "Optional[dict]":
        """Find the dataset identified by the passed dataset_identifier in all passed datasets.

        Args:
            datasets (List[dict]): TIRA representations of datasets.
            dataset_identifier (_type_): identifier of the dataset to be found.

        Returns:
            Optional[dict]: The TIRA representation of the dataset if found in the datasets.
        """
        datasets = [i for i in datasets if "id" in i and len(i["id"]) > 2]

        for dataset in datasets:
            if dataset_identifier and dataset["id"] == dataset_identifier:
                return dataset

        if len(dataset_identifier.split("/")) == 2:
            task_identifier, dataset_in_task = dataset_identifier.split("/")

            for dataset in datasets:
                if "default_task" not in dataset or not dataset["default_task"]:
                    continue

                if not task_identifier or not dataset_in_task:
                    continue

                if task_identifier == dataset["default_task"] and dataset_in_task == dataset["id"]:
                    return dataset

                if task_identifier == dataset["default_task"] and dataset_in_task == dataset["display_name"]:
                    return dataset

        for dataset in datasets:
            if "ir_datasets_id" not in dataset or not dataset["ir_datasets_id"] or len(dataset["ir_datasets_id"]) < 3:
                continue

            if dataset_identifier and dataset_identifier == dataset["ir_datasets_id"]:
                return dataset

        matches = []
        for dataset in datasets:
            if "default_task" not in dataset or not dataset["default_task"]:
                continue

            if dataset_identifier and dataset_identifier == dataset["display_name"]:
                matches.append(dataset)

        if len(matches) == 1:
            return matches[0]



import unittest

from tira.local_client import Client


def retrieval_submission(queries):
    tira = Client("tests/resources/")
    run = tira.trectools.from_submission("ir-benchmarks/tira-ir-starters/retriever", "d1")

    return [
        {"qid": i["query"], "docno": i["docid"], "score": i["score"]}
        for _, i in run.run_data.iterrows()
        if i["query"] in queries
    ]


class TestTrecToolsFromRetrieverSubmissionLoaderTest(unittest.TestCase):
    def test_for_multiple_queries(self):
        expected = [
            {"qid": "1", "docno": "doc-1", "score": 10},
            {"qid": "1", "docno": "doc-2", "score": 9},
            {"qid": "3", "docno": "doc-3", "score": 1},
        ]

        actual = retrieval_submission(["1", "3"])

        self.assertEqual(len(actual), 3)
        self.assertEqual(actual[0], expected[0])
        self.assertEqual(actual[1], expected[1])
        self.assertEqual(actual[2], expected[2])



#!/usr/bin/env python
import argparse
import logging
from pathlib import Path
from platform import python_version
from typing import TYPE_CHECKING, Optional

from tira import __version__
from tira.io_utils import _fmt, log_message, verify_tira_installation
from tira.rest_api_client import Client as RestClient

if TYPE_CHECKING:
    from .tira_client import TiraClient


def download_run(client, approach, dataset):
    help()


"""
Utility functions to setup arguments that are common across multiple commands (e.g., logging).
"""


def setup_logging_args(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "-v",
        "--verbose",
        default=0,
        action="count",
        help="Increases the output verbosity. Default is INFO. Can be repeated for more verbose output.",
    )
    parser.add_argument("-q", "--quiet", action="store_true", help="Disables logging.")


"""
Setup commands for each supported subcommand.
"""


def setup_download_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument(
        "--approach",
        required=False,
        default=None,
        help="Download the outputs of the specified approach. Usage: --approach <task-id>/<user-id>/<approach-name>",
    )
    parser.add_argument("--dataset", required=True, help="The dataset.")
    parser.set_defaults(executable=download_command)


def setup_login_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument("--token", required=True, default=None, help="The token to login to the server.")
    parser.add_argument(
        "--print-docker-auth",
        required=False,
        default=False,
        action="store_true",
        help="Print the docker credentials as TIRA_DOCKER_REGISTRY_USER and TIRA_DOCKER_REGISTRY_TOKEN to stdout.",
    )
    parser.set_defaults(executable=login_command)


def setup_evaluation_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument(
        "--directory", required=True, default=None, help="The directory with the predictions to evaluate."
    )
    parser.add_argument("--dataset", required=True, help="The dataset for which the predictions are made.")
    parser.set_defaults(executable=evaluate_command)


def evaluate_command(directory: Path, dataset: str, **kwargs) -> int:
    client: "TiraClient" = RestClient()
    client.evaluate(Path(directory), dataset)

    return 0


def setup_code_submission_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.add_argument(
        "--path",
        required=True,
        default=None,
        help="The path used to build the docker image, must be in a clean git repository.",
    )
    parser.add_argument(
        "--command",
        required=False,
        default=None,
        help="The command that executes your approach in the docker image (default: the Docker Entrypoint). The command should read the data from $inputDataset to write predictions to $outputDir.",
    )
    parser.add_argument(
        "--dry-run",
        required=False,
        default=False,
        action="store_true",
        help="Make a dry-run, i.e., to develop your solution, i.e., not uploading to TIRA and not requiring that the git repo is clean.",
    )
    parser.add_argument(
        "--allow-network",
        required=False,
        default=False,
        action="store_true",
        help="Allow network communication. Within TIRA itself, software is executed in a sandbox without internet access",
    )
    parser.add_argument("--task", required=True, default=None, help="The task to which the code should be submitted.")

    parser.set_defaults(executable=code_submission_command)


def setup_verify_installation(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.set_defaults(executable=verify_installation_command)


def setup_upload_command(parser: argparse.ArgumentParser) -> None:
    setup_logging_args(parser)
    parser.set_defaults(executable=upload_command)


"""
Implementations of each command. They don't all need to be here but if your command is short and sweet and you don't
don't know, where else to put it, this is a good place.
"""


def download_command(dataset: str, approach: Optional[str] = None, **kwargs) -> int:
    client: "TiraClient" = RestClient()
    if approach is not None:
        print(client.get_run_output(approach, dataset))
    else:
        print(client.download_dataset(None, dataset))
    return 0


def login_command(token: str, print_docker_auth: bool, **kwargs) -> int:
    client: "TiraClient" = RestClient()
    client.login(token)
    if print_docker_auth:
        print(client.local_execution.docker_client_is_authenticated())
    return 0


def code_submission_command(
    path: Path, task: str, dry_run: bool, allow_network: bool, command: Optional[str], **kwargs
) -> int:
    client: "TiraClient" = RestClient()
    client.submit_code(Path(path), task, command, dry_run=dry_run, allow_network=allow_network)

    return 0


def verify_installation_command(**kwards) -> int:
    status = verify_tira_installation()

    print("\nResult:")
    msg = "Your TIRA installation is valid."
    if status == _fmt.WARN:
        msg = "There are some warnings, you might not use all features."

    if status == _fmt.ERROR:
        msg = "Your installation is not valid, you might not use all features."

    log_message(msg, status)
    return 0 if status == _fmt.OK else 1


def upload_command() -> None:
    pass


"""
Finally, parsing the arguments and running the chose command.
"""


def parse_args():
    parser = argparse.ArgumentParser(prog="tira-cli")
    version_str = f"TIRA v{__version__} using Python v{python_version()}"
    parser.add_argument("-v", "--version", action="version", version=version_str)

    # Register all subcommands here:
    subparsers = parser.add_subparsers(dest="command", required=True)
    setup_download_command(subparsers.add_parser("download", help="Download runs or datasets from TIRA.io"))
    setup_upload_command(subparsers.add_parser("upload", help="Upload runs or datasets to TIRA.io"))
    setup_login_command(subparsers.add_parser("login", help="Login your TIRA client to the tira server."))
    setup_verify_installation(
        subparsers.add_parser("verify-installation", help="Verify that your local TIRA client is correctly installed.")
    )
    setup_code_submission_command(
        subparsers.add_parser("code-submission", help="Make a code submission via Docker from a git repository.")
    )
    setup_evaluation_command(subparsers.add_parser("evaluate", help="Evaluate some predictions on your local machine."))

    return parser.parse_args()


def main() -> int:
    args = parse_args()
    if not hasattr(args, "executable") or args.executable is None:
        logging.critical("No subcommand was found or it was not set up correctly.")
        logging.critical(
            f"Command: {vars(args).get('command', None)}; Executable: {vars(args).get('executable', None)}"
        )
        return -1

    # Potential bug: a user could supply his own "quiet" argument without using setup_logging_args and we would handle
    # it as if it was a logging argument, which it may not be. Probably the best solution right now would be to create a
    # custom argparse.Action for setup_logging_args to handle the --quiet and --verbose options... but that would be a
    # bit more effort and we can tackle that when this actually clashes (i.e., if we need --quiet or --verbose handled
    # differently in another command.
    if hasattr(args, "quiet") and args.quiet:
        logging.disable()
    else:
        if not hasattr(args, "verbose"):
            args.verbose = 0
        verbosity_levels = [logging.CRITICAL, logging.ERROR, logging.WARN, logging.INFO, logging.DEBUG]
        default_verbosity = verbosity_levels.index(logging.INFO)
        verbosity = verbosity_levels[min(default_verbosity + args.verbose, len(verbosity_levels) - 1)]
        logging.basicConfig(level=verbosity)

    # Subcommands store their executable into args.executable which takes the parsed arguments and returns an integer.
    # We just dump all the arguments into the function and let it sort it out itself.
    return args.executable(**vars(args))



import gzip
import json
import os
import re
import xml.dom.minidom
from enum import Enum
from glob import glob
from pathlib import Path
from typing import Sequence, Union

import yaml


class FormatMsgType(Enum):
    OK = 0
    WARN = 1
    ERROR = 2


_fmt = FormatMsgType


def log_message(message: str, level: _fmt):
    """
    Prints a formatted log message with a symbol indicating the status.

    Parameters:
    - message (str): The log message to display.
    - level (_fmt): The level of the message; can be _fmt.OK, _fmt.WARN, _fmt.ERROR.
    """
    symbols = {
        _fmt.OK: "\033[92m\u2713\033[0m",  # Green check mark
        _fmt.WARN: "\033[93m" + b"\xe2\x9a\xa0".decode("utf-8") + "\033[0m",  # Yellow warning
        _fmt.ERROR: "\033[91m" + b"\xe2\x9c\x96".decode("utf-8") + "\033[0m",  # Red cross
    }

    symbol = symbols[level]
    print(f"{symbol} {message}")


class FormatBase:
    def all_lines(self, f: Path):
        try:
            f_size = f.stat().st_size
        except:
            raise ValueError(f"File {f} is not readable.")

        if f_size > self.max_size():
            raise ValueError(f"File {f} is too large (size {f_size} exceeds configured {self.max_size()}).")

        if str(f).endswith(".gz"):
            with gzip.open(f, "rt") as o:
                return o.readlines()
        else:
            with open(f, "r") as o:
                return o.readlines()

    def max_size(self):
        return 25 * 1024 * 1024


class RunFormat(FormatBase):
    """Checks if a given output is a valid run file."""

    def check_format(self, run_output: Path):
        if (run_output / "run.txt").exists() and (run_output / "run.txt.gz").exists():
            msg = f"Found multiple run.txt or run.txt.gz files: {os.listdir(run_output)} ."
            return [_fmt.ERROR, msg]
        if not (run_output / "run.txt").exists() and not (run_output / "run.txt.gz").exists():
            msg = "No file run.txt or run.txt.gz was found, only the files "
            msg += str(os.listdir(run_output)) + " were available."
            return [_fmt.ERROR, msg]

        try:
            # maximum size of 10MB
            lines = self.all_lines(run_output)
        except Exception as e:
            return [_fmt.ERROR, repr(e)]

        if len(lines) < 10:
            return [_fmt.ERROR, f"The run file contains only {len(lines)} lines, this is likely an error."]

        query_to_docs = {}
        for l in lines:
            l = l.strip()
            cols = re.split("\\s+", l)
            if len(cols) != 6:
                return [
                    _fmt.ERROR,
                    f'Invalid line in the run file, expected 6 columns, but found a line "{l}" with {len(cols)} columns.',
                ]
            qid, _, docno, _, _, _ = cols
            if qid not in query_to_docs:
                query_to_docs[qid] = set()
            if docno in query_to_docs[qid]:
                return [
                    _fmt.ERROR,
                    f'The run file has duplicate documents: the document with id "{docno}" appears multiple times for query "{qid}".',
                ]
            query_to_docs[qid].add(docno)

        if len(query_to_docs.keys()) < 3:
            return [_fmt.ERROR, f"The run file has only {len(query_to_docs)} queries which is likely an error."]

        return [_fmt.OK, "The run.txt file has the correct format."]

    def all_lines(self, run_output):
        if (run_output / "run.txt").exists():
            return [i.strip() for i in super().all_lines(run_output / "run.txt")]
        if (run_output / "run.txt.gz").exists():
            return [i.strip() for i in super().all_lines(run_output / "run.txt.gz")]
        else:
            raise ValueError("Could not find a file run.txt or run.txt.gz")


class JsonlFormat(FormatBase):
    """Checks if a given output is a valid jsonl file."""

    def __init__(self, required_fields=("id",), minimum_lines=3):
        self.required_fields = required_fields
        self.minimum_lines = minimum_lines

    def check_format(self, run_output: Path):
        try:
            lines = self.all_lines(run_output)

            if len(lines) < self.minimum_lines:
                return [_fmt.ERROR, f"The *.jsonl file contains only {len(lines)} lines, this is likely an error."]

            return [_fmt.OK, "The jsonl file has the correct format."]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

    def fail_if_json_line_is_not_valid(self, line):
        for field in self.required_fields:
            if not line or field not in line:
                raise ValueError(f'The json line misses the required field "{field}": "{json.dumps(line)}".')

    def all_lines(self, run_output):
        if (str(run_output).endswith(".jsonl") or str(run_output).endswith(".jsonl.gz")) and run_output.is_file():
            matches = [run_output]
        else:
            matches = [
                run_output / i for i in os.listdir(run_output) if i.endswith(".jsonl") or i.endswith(".jsonl.gz")
            ]

        if len(matches) != 1:
            raise ValueError(
                "No unique *.jsonl file was found, only the files " + str(os.listdir(run_output)) + " were available."
            )

        ret_raw = [i.strip() for i in super().all_lines(matches[0])]
        ret = []
        for i in ret_raw:
            try:
                ret.append(json.loads(i))
            except:
                raise ValueError(f'The file {matches[0]} contains a line that could not be parsed: "{i}".')

        for i in ret:
            self.fail_if_json_line_is_not_valid(i)

        return ret


class GenIrSimulationFormat(JsonlFormat):
    def __init__(self):
        super().__init__(required_fields=(), minimum_lines=1)

    def fail_if_json_line_is_not_valid(self, line):
        if "simulation" not in line:
            raise ValueError('The json line misses the required field "simulation".')
        line = line["simulation"]

        if "configuration" not in line:
            raise ValueError('The json line misses the required field "simulation.configuration".')

        if "userTurns" not in line:
            raise ValueError('The json line misses the required field "simulation.userTurns".')


class TsvFormat(FormatBase):
    """Checks if a given output is a valid tsv file."""

    def check_format(self, run_output: Path):
        try:
            lines = self.all_lines(run_output)
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < 5:
            return [_fmt.ERROR, f"The *.tsv file contains only {len(lines)} lines, this is likely an error."]

        if len(lines[0]) < 2:
            return [_fmt.ERROR, "The *.tsv file contains only one column, this is likely an error."]

        return [_fmt.OK, "The tsv file has the correct format."]

    def all_lines(self, run_output):
        matches = [i for i in os.listdir(run_output) if i.endswith(".tsv")]
        if len(matches) != 1:
            msg = "No unique *.tsv file was found, only the files "
            msg += str(os.listdir(run_output)) + " were available."
            raise ValueError(msg)

        f = run_output / matches[0]
        with open(f, "r") as tsv_file:
            columns = None
            ret = []
            for l in tsv_file:
                l_parsed = l.strip().split("\t")
                if columns is None:
                    columns = len(l_parsed)
                if len(l_parsed) != columns:
                    raise ValueError("The *.tsv file is invalid: The number of columns varies.")
                ret += [l_parsed]
            return ret


class TextAlignmentFeaturesFormat(FormatBase):
    """
    The code was refactored from https://github.com/pan-webis-de/pan-code/blob/master/sepln09/pan09-plagiarism-detection-performance-measures.py#L276
    """

    def check_format(self, run_output: Path):
        try:
            lines = self.all_lines(run_output)
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < 3:
            return [
                _fmt.ERROR,
                f"The text-alignment-feature directory contains only {len(lines)} instances, this is likely an error.",
            ]

        return [_fmt.OK, "The directory has the correct format."]

    def all_lines(self, run_output):
        return self.extract_annotations_from_files(run_output, ["plagiarism", "detected-plagiarism"])

    def extract_annotations_from_file(self, xmlfile, tagnames):
        """Returns a set of plagiarism annotations from an XML file."""
        doc = xml.dom.minidom.parse(xmlfile)
        annotations = []
        if not doc.documentElement.hasAttribute("reference"):
            return annotations
        t_ref = doc.documentElement.getAttribute("reference")
        for node in doc.documentElement.childNodes:
            for tagname in tagnames:
                if (
                    node.nodeType == xml.dom.Node.ELEMENT_NODE
                    and node.hasAttribute("name")
                    and node.getAttribute("name").endswith(tagname)
                ):
                    ann = self.extract_annotation_from_node(node, t_ref)
                    if ann:
                        ann["type"] = tagname
                        annotations.append(ann)
        return annotations

    def extract_annotation_from_node(self, xmlnode, t_ref):
        """Returns a plagiarism annotation from an XML feature tag node."""
        if not (xmlnode.hasAttribute("this_offset") and xmlnode.hasAttribute("this_length")):
            return False
        t_off = int(xmlnode.getAttribute("this_offset"))
        t_len = int(xmlnode.getAttribute("this_length"))
        s_ref, s_off, s_len, ext = "", 0, 0, False
        if (
            xmlnode.hasAttribute("source_reference")
            and xmlnode.hasAttribute("source_offset")
            and xmlnode.hasAttribute("source_length")
        ):
            s_ref = xmlnode.getAttribute("source_reference")
            s_off = int(xmlnode.getAttribute("source_offset"))
            s_len = int(xmlnode.getAttribute("source_length"))
            ext = True
        return {
            "this_reference": t_ref,
            "this_offset": t_off,
            "this_length": t_len,
            "source_reference": s_ref,
            "source_offset": s_off,
            "source_length": s_len,
            "is_external": ext,
        }

    def extract_annotations_from_files(self, path, tagnames):
        """Returns a set of plagiarism annotations from XML files below path."""
        if not os.path.exists(path):
            raise ValueError(f"Path not accessible: {path}")

        annotations = []
        xmlfiles = glob(os.path.join(path, "*.xml"))
        xmlfiles.extend(glob(os.path.join(path, os.path.join("*", "*.xml"))))
        for xmlfile in xmlfiles:
            annotations += self.extract_annotations_from_file(xmlfile, tagnames)
        return annotations


class PanStyleChangeDetectionCorpusFormat(FormatBase):
    """Checks if a given directory contains a valid style-change-detection PAN corpus. An example is located here: https://pan.webis.de/clef24/pan24-web/style-change-detection.html"""

    def check_format(self, run_output: Path):
        try:
            lines = [i for i in self.yield_lines(run_output, False)]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < 3:
            return [
                _fmt.ERROR,
                f"The style-change-detection directory contains only {len(lines)} instances, this is likely an error.",
            ]

        return [_fmt.OK, "The directory has the correct format."]

    def all_lines(self, run_output):
        return [i for i in self.yield_lines(run_output, True)]

    def yield_lines(self, directory, load_content):
        matches = (
            glob(f"{directory}/problem-*.txt")
            + glob(f"{directory}/*/problem-*.txt")
            + glob(f"{directory}/*/*/problem-*.txt")
            + glob(f"{directory}/*/*/*/problem-*.txt")
        )

        for i in matches:
            entry = {"id": i.split("/")[-1].replace(".txt", "")}

            if load_content:
                with open(i, "r", newline="") as f:
                    entry["text"] = f.read()

            yield entry


class PanStyleChangeDetectionPredictionFormat(FormatBase):
    """Checks if a given directory contains a valid style-change-detection PAN predictions. An example is located here: https://pan.webis.de/clef24/pan24-web/style-change-detection.html"""

    def check_format(self, run_output: Path):
        try:
            lines = [i for i in self.yield_lines(run_output, False)]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < 3:
            return [
                _fmt.ERROR,
                f"The style-change-detection directory contains only {len(lines)} instances, this is likely an error.",
            ]

        return [_fmt.OK, "The directory has the correct format."]

    def all_lines(self, run_output):
        return [i for i in self.yield_lines(run_output, True)]

    def yield_lines(self, directory, load_content):
        matches = (
            glob(f"{directory}/*-problem-*.json")
            + glob(f"{directory}/*/*-problem-*.json")
            + glob(f"{directory}/*/*/*-problem-*.json")
            + glob(f"{directory}/*/*/*/*-problem-*.json")
        )

        for i in matches:
            prediction_type, problem_id = i.split("/")[-1].replace(".json", "").split("-problem-")
            entry = {"problem": problem_id, "type": prediction_type}

            if load_content:
                with open(i, "r", newline="") as f:
                    entry.update(json.loads(f.read()))

            yield entry


class TextAlignmentCorpusFormat(FormatBase):
    """Checks if a given directory contains a valid PAN text alignment corpus. The code was refactored from https://github.com/pan-webis-de/pan-code/blob/master/clef12/text-alignment/pan12-text-alignment-baseline.py#L202"""

    def check_format(self, run_output: Path):
        try:
            lines = [i for i in self.yield_lines(run_output, False)]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

        if len(lines) < 3:
            return [
                _fmt.ERROR,
                f"The text-alignment directory contains only {len(lines)} instances, this is likely an error.",
            ]

        return [_fmt.OK, "The directory has the correct format."]

    def all_lines(self, run_output):
        return [i for i in self.yield_lines(run_output, True)]

    def yield_lines(self, run_output, load_content):
        matches = [i for i in os.listdir(run_output) if i.endswith("pairs")]
        if len(matches) != 1:
            msg = "No unique pair file was found, only the files "
            msg += str(os.listdir(run_output)) + " were available."
            raise ValueError(msg)

        f = run_output / matches[0]
        with open(f, "r") as tsv_file:
            for l in tsv_file:
                l_parsed = l.strip().split()
                if len(l_parsed) != 2:
                    raise ValueError("The pair file is invalid: I expect always two lines per entry.")
                entry = {}
                for k, v in zip(l_parsed, ["suspicious", "source"]):
                    entry[f"{v}_document_id"] = k.replace(".txt", "")
                    matches = (
                        glob(f"{run_output}/{k}")
                        + glob(f"{run_output}/{k}.txt")
                        + glob(f"{run_output}/*/{k}")
                        + glob(f"{run_output}/*/{k}.txt")
                    )

                    if len(matches) != 1:
                        msg = f"No unique text file was found for id {k}. Only "
                        msg += str(os.listdir(run_output)) + " were available."

                        raise ValueError(msg)
                    if load_content:
                        with open(matches[0]) as f:
                            entry[f"{v}_document_text"] = f.read()

                yield entry


class IrMetadataFormat(FormatBase):
    """Checks if a given output contains valid ir_metadata."""

    def __init__(self, positive_fields=("platform", "implementation", "resources")):
        self.positive_fields = positive_fields

    def check_format(self, run_output: Path):
        try:
            lines = list(self.yield_lines(run_output))

            if len(lines) < 1:
                return [_fmt.ERROR, "At least one valid ir_metadata file was expected, but there was none."]

            return [_fmt.OK, "The output provides valid ir_metadata."]
        except Exception as e:
            return [_fmt.ERROR, str(e)]

    def all_lines(self, run_output):
        return [i for i in self.yield_lines(run_output)]

    def yield_lines(self, run_output: Path):
        candidates = [run_output]
        for pattern in ["/*.yml", "/*.yaml", "/.*.yml", "/.*.yaml"]:
            for depth in ["", "/**", "/**/**"]:
                candidates += glob(str(run_output) + depth + pattern)

        for candidate in candidates:
            if not Path(candidate).exists() or not Path(candidate).is_file():
                continue

            with open(candidate) as stream:
                try:
                    content = yaml.safe_load(stream)
                    at_least_one_positive_field = any(i in content for i in self.positive_fields)

                    if not at_least_one_positive_field:
                        continue

                    yield {"name": candidate.split("/")[-1], "content": content}
                except yaml.YAMLError:
                    pass


FORMAT_TO_CHECK = {
    "run.txt": lambda: RunFormat(),
    "*.jsonl": lambda: JsonlFormat(),
    "*.tsv": lambda: TsvFormat(),
    "text-alignment-corpus": lambda: TextAlignmentCorpusFormat(),
    "text-alignment-features": lambda: TextAlignmentFeaturesFormat(),
    "style-change-detection-corpus": lambda: PanStyleChangeDetectionCorpusFormat(),
    "style-change-detection-predictions": lambda: PanStyleChangeDetectionPredictionFormat(),
    "GenIR-Simulation": lambda: GenIrSimulationFormat(),
    "ir_metadata": lambda: IrMetadataFormat(),
}

SUPPORTED_FORMATS = set(sorted(list(FORMAT_TO_CHECK.keys())))


def lines_if_valid(run_output: Path, format: Union[str, Sequence[str]]):
    """Load all lines from a user file if they are valid

    Args:
        format (Union[str, Sequence[str]]): The allowed format or a list of allowed formats.
        run_output (Path): the output produced by some run that is to-be checked.
    """
    if format not in SUPPORTED_FORMATS:
        raise ValueError(f"Format {format} is not supported. Supported formats are {SUPPORTED_FORMATS}.")

    checker = FORMAT_TO_CHECK[format]()
    result, msg = checker.check_format(run_output)
    if result != _fmt.OK:
        raise ValueError(msg)

    return checker.all_lines(run_output)


def check_format(run_output: Path, format: Union[str, Sequence[str]]):
    """Check if the provided run output is in the specified format. Provides debug messages intended for users.

    Args:
        format (Union[str, Sequence[str]]): The allowed format or a list of allowed formats.
        run_output (Path): the output produced by some run that is to-be checked.
    """
    if format not in SUPPORTED_FORMATS:
        raise ValueError(f"Format {format} is not supported. Supported formats are {SUPPORTED_FORMATS}.")

    checker = FORMAT_TO_CHECK[format]()
    return checker.check_format(run_output)



import unittest

import pandas as pd

from tira.check_format import lines_if_valid
from tira.serialize_format import serialize_in_temporary_directory

from ..format_check import VALID_RUN_OUTPUT

VALID_TRECTOOLS_RUN = pd.DataFrame(
    [
        {"query": 1, "q0": "Q0", "docid": "d-1", "rank": 1, "score": 10, "system": "s"},
        {"query": 1, "q0": "Q0", "docid": "d-2", "rank": 2, "score": 9, "system": "s"},
        {"query": 1, "q0": "Q0", "docid": "d-3", "rank": 3, "score": 8, "system": "s"},
        {"query": 1, "q0": "Q0", "docid": "d-4", "rank": 4, "score": 7, "system": "s"},
        {"query": 2, "q0": "Q0", "docid": "d-1", "rank": 1, "score": 10, "system": "s"},
        {"query": 2, "q0": "Q0", "docid": "d-2", "rank": 2, "score": 9, "system": "s"},
        {"query": 2, "q0": "Q0", "docid": "d-3", "rank": 3, "score": 8, "system": "s"},
        {"query": 2, "q0": "Q0", "docid": "d-4", "rank": 4, "score": 7, "system": "s"},
        {"query": 3, "q0": "Q0", "docid": "d-1", "rank": 1, "score": 10, "system": "s"},
        {"query": 3, "q0": "Q0", "docid": "d-2", "rank": 2, "score": 9, "system": "s"},
        {"query": 3, "q0": "Q0", "docid": "d-3", "rank": 3, "score": 8, "system": "s"},
        {"query": 3, "q0": "Q0", "docid": "d-4", "rank": 4, "score": 7, "system": "s"},
    ]
)


VALID_PYTERRIER_RUN = pd.DataFrame(
    [
        {"qid": 1, "q0": "Q0", "docno": "d-1", "rank": 1, "score": 10, "system": "s"},
        {"qid": 1, "q0": "Q0", "docno": "d-2", "rank": 2, "score": 9, "system": "s"},
        {"qid": 1, "q0": "Q0", "docno": "d-4", "rank": 4, "score": 7, "system": "s"},
        {"qid": 2, "q0": "Q0", "docno": "d-1", "rank": 1, "score": 10, "system": "s"},
        {"qid": 2, "q0": "Q0", "docno": "d-2", "rank": 2, "score": 9, "system": "s"},
        {"qid": 2, "q0": "Q0", "docno": "d-3", "rank": 3, "score": 8, "system": "s"},
        {"qid": 2, "q0": "Q0", "docno": "d-4", "rank": 4, "score": 7, "system": "s"},
        {"qid": 3, "q0": "Q0", "docno": "d-1", "rank": 1, "score": 10, "system": "s"},
        {"qid": 3, "q0": "Q0", "docno": "d-2", "rank": 2, "score": 9, "system": "s"},
        {"qid": 3, "q0": "Q0", "docno": "d-3", "rank": 3, "score": 8, "system": "s"},
        {"qid": 3, "q0": "Q0", "docno": "d-4", "rank": 4, "score": 7, "system": "s"},
    ]
)


class TrecToolsRunMock:
    def __init__(self, run):
        self.run_data = run


class TestSerializationOfRunFiles(unittest.TestCase):
    def test_for_valid_run_file_as_path(self):
        input_file = VALID_RUN_OUTPUT / "run.txt"
        expected = 10
        actual_dir = serialize_in_temporary_directory(input_file, "run.txt")
        lines = lines_if_valid(actual_dir, "run.txt")

        self.assertEqual(expected, len(lines))

    def test_for_trectools_run(self):
        expected = 12
        actual_dir = serialize_in_temporary_directory(VALID_TRECTOOLS_RUN, "run.txt")
        lines = lines_if_valid(actual_dir, "run.txt")

        self.assertEqual(expected, len(lines))

    def test_for_pyterrier_run(self):
        expected = 11
        actual_dir = serialize_in_temporary_directory(VALID_PYTERRIER_RUN, "run.txt")
        lines = lines_if_valid(actual_dir, "run.txt")

        self.assertEqual(expected, len(lines))

    def test_for_trectools_run_in_rundata(self):
        expected = 12
        actual_dir = serialize_in_temporary_directory(TrecToolsRunMock(VALID_TRECTOOLS_RUN), "run.txt")
        lines = lines_if_valid(actual_dir, "run.txt")

        self.assertEqual(expected, len(lines))

    def test_for_pyterrier_run_in_rundata(self):
        expected = 11
        actual_dir = serialize_in_temporary_directory(TrecToolsRunMock(VALID_PYTERRIER_RUN), "run.txt")
        lines = lines_if_valid(actual_dir, "run.txt")

        self.assertEqual(expected, len(lines))

    def test_for_valid_run_file_as_str(self):
        input_file = str(VALID_RUN_OUTPUT / "run.txt")
        expected = 10
        actual_dir = serialize_in_temporary_directory(input_file, "run.txt")
        lines = lines_if_valid(actual_dir, "run.txt")

        self.assertEqual(expected, len(lines))

    def test_for_non_existing_file_as_path(self):
        input_file = VALID_RUN_OUTPUT / "does-not-exist"

        with self.assertRaises(Exception) as context:
            serialize_in_temporary_directory(input_file, "run.txt")

        self.assertTrue("The passed run file" in str(repr(context.exception)))

    def test_for_invalid_trectools_dataframe(self):
        input_dataframe = VALID_TRECTOOLS_RUN.copy()
        del input_dataframe["query"]

        with self.assertRaises(Exception) as context:
            serialize_in_temporary_directory(input_dataframe, "run.txt")

        self.assertTrue("The passed run is missing the required columns " in str(repr(context.exception)))

    def test_for_invalid_pyterrier_dataframe(self):
        input_dataframe = VALID_PYTERRIER_RUN.copy()
        del input_dataframe["qid"]

        with self.assertRaises(Exception) as context:
            serialize_in_temporary_directory(input_dataframe, "run.txt")

        self.assertTrue("The passed run is missing the required columns " in str(repr(context.exception)))



import json
import logging
import os
import shlex
from pathlib import Path

from tira.io_utils import all_lines_to_pandas


def ensure_pyterrier_is_loaded(
    boot_packages=("com.github.terrierteam:terrier-prf:-SNAPSHOT",), packages=(), patch_ir_datasets=True
):
    import pyterrier as pt

    # Detect if we are in the TIRA sandbox
    if patch_ir_datasets:
        try:
            import ir_datasets as original_ir_datasets

            original_ir_datasets.load = load_ir_datasets().load
            logging.info(
                "Due to execution in TIRA, I have patched ir_datasets to always return the single input dataset mounted"
                " to the sandbox."
            )
        except Exception as e:
            logging.error("Could not patch ir_datasets.", exc_info=e)
    else:
        import ir_datasets as original_ir_datasets

        from tira.ir_datasets_util import original_ir_datasets_load

        original_ir_datasets.load = original_ir_datasets_load

    pt_version = os.environ.get("PYTERRIER_VERSION", "5.7")
    pt_helper_version = os.environ.get("PYTERRIER_HELPER_VERSION", "0.0.7")

    if not pt.started():
        logging.info(f"Start PyTerrier with version={pt_version}, helper_version={pt_helper_version}, no_download=True")
        pt.init(
            version=pt_version,
            helper_version=pt_helper_version,
            no_download=True,
            boot_packages=list(boot_packages),
            packages=list(packages),
        )


def get_preconfigured_chatnoir_client(
    config_directory, features=["TARGET_URI"], num_results=10, retries=25, page_size=10
):
    from chatnoir_api import Index as ChatNoirIndex
    from chatnoir_pyterrier import ChatNoirRetrieve
    from chatnoir_pyterrier.feature import Feature

    chatnoir_config = json.load(open(config_directory + "/chatnoir-credentials.json"))

    chatnoir = ChatNoirRetrieve(api_key=chatnoir_config["apikey"], staging=chatnoir_config.get("staging", False))
    chatnoir.features = [getattr(Feature, i) for i in features]
    chatnoir.num_results = num_results
    chatnoir.retries = retries
    chatnoir.page_size = page_size
    chatnoir.index = getattr(ChatNoirIndex, chatnoir_config["index"])

    logging.info(
        f"ChatNoir Client will retrieve the top-{chatnoir.num_results} with page size of {chatnoir.page_size} from"
        f" index {chatnoir_config['index']} with {chatnoir.retries} retries."
    )

    return chatnoir


def get_output_directory(default_output: str = "/tmp/"):
    output_directory = os.environ.get("TIRA_OUTPUT_DIR", default_output)
    logging.info(f"The output directory is {output_directory}")
    return output_directory


def get_input_directory_and_output_directory(default_input, default_output: str = "/tmp/"):
    input_directory = os.environ.get("TIRA_INPUT_DATASET", None)

    if input_directory:
        print(f"I will read the input data from {input_directory}.")
    else:
        input_directory = default_input
        print(f"I will use a small hardcoded example located in {input_directory}.")

    return (input_directory, get_output_directory(default_output))


def is_running_as_inference_server():
    return os.environ.get("TIRA_INFERENCE_SERVER", None) is not None


def load_rerank_data(default, load_default_text=True):
    default_input = get_input_directory_and_output_directory(default)[0]

    if not os.path.isdir(default_input) and len(default.split("/")) == 2:
        from tira.rest_api_client import Client as RestClient

        default_input = RestClient().download_dataset(default.split("/")[0], default.split("/")[1])

    if not default_input.endswith("rerank.jsonl") and not default_input.endswith("rerank.jsonl.gz"):
        if os.path.isfile(default_input + "/rerank.jsonl.gz"):
            default_input = default_input + "/rerank.jsonl.gz"
        elif os.path.isfile(default_input + "/rerank.jsonl"):
            default_input = default_input + "/rerank.jsonl"

    return all_lines_to_pandas(default_input, load_default_text)


def register_rerank_data_to_ir_datasets(path_to_rerank_file, ir_dataset_id, original_ir_datasets_id=None):
    """
    Load a dynamic ir_datasets integration from a given re_rank_file.
    The dataset will be registered for the id ir_dataset_id.
    The original_ir_datasets_id is used to infer the class of documents, qrels, and queries.
    """
    from tira.ir_datasets_util import register_dataset_from_re_rank_file

    default_input = get_input_directory_and_output_directory(path_to_rerank_file)[0]

    if not default_input.endswith("rerank.jsonl") and not default_input.endswith("rerank.jsonl.gz"):
        if os.path.isfile(default_input + "/rerank.jsonl.gz"):
            default_input = default_input + "/rerank.jsonl.gz"
        elif os.path.isfile(default_input + "/rerank.jsonl"):
            default_input = default_input + "/rerank.jsonl"

    register_dataset_from_re_rank_file(ir_dataset_id, default_input, original_ir_datasets_id)


def persist_and_normalize_run(
    run, system_name, default_output=None, output_file=None, depth=1000, upload_to_tira=None, tira_client=None
):
    if output_file is None and default_output is None:
        print(
            'I use the environment variable "TIRA_OUTPUT_DIR" to determine where I should store the run file using "."'
            " as default."
        )
        output_file = os.environ.get("TIRA_OUTPUT_DIR", ".")

    if default_output is not None:
        if os.environ.get("TIRA_OUTPUT_DIR") is None:
            print(f'The run file is normalized outside the TIRA sandbox, I will store it at "{default_output}".')
            output_file = default_output
        else:
            output_file = os.environ.get("TIRA_OUTPUT_DIR")
            print(f'The run file is normalized inside the TIRA sandbox, I will store it at "{output_file}".')

    if not output_file.endswith("run.txt"):
        output_file = output_file + "/run.txt"
    if upload_to_tira and not in_tira_sandbox():
        tira = _tira_client(tira_client)
        tmp = tira.get_dataset(upload_to_tira)
        if not tmp or "dataset_id" not in tmp:
            upload_to_tira = None
    else:
        upload_to_tira = None

    if upload_to_tira and tira:
        output_file = output_file + ".gz"
    normalize_run(run, system_name, depth).to_csv(output_file, sep=" ", header=False, index=False)

    print(f'Done. run file is stored under "{output_file}".')

    if upload_to_tira and tira:
        output_file = Path(output_file).parent
        upload_run_anonymous(output_file, tira, upload_to_tira)


def _tira_client(default_tira_client=None):
    if in_tira_sandbox():
        return None

    from tira.rest_api_client import Client as RestClient

    if default_tira_client:
        return default_tira_client
    else:
        return RestClient()


def upload_run_anonymous(directory: Path = None, tira_client=None, dataset_id=None):
    tira = _tira_client(tira_client)
    if not tira:
        return

    upload_to_tira = tira.get_dataset(dataset_id)
    tira.upload_run_anonymous(directory, upload_to_tira["dataset_id"])


def temporary_directory():
    import tempfile

    try:
        ret = tempfile.TemporaryDirectory(prefix="tira-", delete=False)
        ret = Path(ret.name)
    except:
        ret = tempfile.TemporaryDirectory(prefix="tira-")
        ret = Path(ret.name)

    ret.mkdir(parents=True, exist_ok=True)
    return ret


def normalize_run(run, system_name, depth=1000):
    try:
        run["qid"] = run["qid"].astype(int)
    except Exception:
        pass

    run["system"] = system_name
    run = run.copy().sort_values(["qid", "score", "docno"], ascending=[True, False, False]).reset_index()

    if "Q0" not in run.columns:
        run["Q0"] = 0

    run = run.groupby("qid")[["qid", "Q0", "docno", "score", "system"]].head(depth)

    # Make sure that rank position starts by 1
    run["rank"] = 1
    run["rank"] = run.groupby("qid")["rank"].cumsum()

    return run[["qid", "Q0", "docno", "rank", "score", "system"]]


def extract_to_be_executed_notebook_from_command_or_none(command: str):
    command = command.replace(";", " ").replace("&", " ").replace("|", " ") if command is not None else None
    if command is not None and "--notebook" in command:
        return command.split("--notebook")[1].strip().split(" ")[0].strip()

    if command is not None and ".py" in command:
        for arg in shlex.split(command, posix=False):
            if arg.endswith(".py") or arg.endswith('.py"') or arg.endswith(".py'"):
                return arg

    if command is not None:
        command = shlex.split(command, posix=False)
        if len(command) > 0:
            command = command[0]

        if command and (
            command.endswith(".sh")
            or command.endswith(".bash")
            or command.endswith('.sh"')
            or command.endswith(".sh'")
            or command.endswith('.bash"')
            or command.endswith(".bash'")
        ):
            return command

    return None


def extract_ast_value(v):
    if hasattr(v, "s"):
        # python3.7
        return v.s
    if hasattr(v, "n"):
        # python3.7
        return v.n
    else:
        return v.value


def parse_ast_extract_assignment(python_line: str):
    try:
        import ast

        python_line = ast.parse(python_line).body[0]

        return python_line.targets[0].id, extract_ast_value(python_line.value)
    except Exception:
        return None, None


def parse_extraction_of_tira_approach_bash(bash_line: str):
    try:
        bash_line = bash_line.split("tira-cli")[1]
        bash_line = bash_line.split("download")[1]
        bash_line = bash_line.split("--approach")[1].strip()
        if bash_line.startswith("'"):
            return bash_line[1:].split("'")[0]
        if bash_line.startswith('"'):
            return bash_line[1:].split('"')[0]

        return bash_line.split(")")[0].split()[0].strip()
    except Exception:
        return None


def parse_extraction_of_tira_approach(python_line: str):
    try:
        import ast

        python_line = ast.parse(python_line.split("%")[0].strip()).body[0]

        if "op" in dir(python_line.value) and "left" in dir(python_line.value) and "right" in dir(python_line.value):
            python_line.value = python_line.value.left

        if "attr" in dir(python_line.value.func.value) and python_line.value.func.value.attr != "pt":
            return None

        if not ("id" in dir(python_line.value.func.value) and python_line.value.func.value.id == "tira") and not (
            "value" in dir(python_line.value.func.value) and python_line.value.func.value.value.id == "tira"
        ):
            return None

        if (
            python_line.value.func.attr != "index"
            and python_line.value.func.attr != "from_submission"
            and python_line.value.func.attr != "get_run_output"
            and python_line.value.func.attr != "transform_queries"
            and python_line.value.func.attr != "transform_documents"
        ):
            return None

        return extract_ast_value(python_line.value.args[0])
    except Exception:
        return None


def __read_src_lines(notebook: Path):
    if not notebook.exists():
        return []

    ret = []
    if notebook.suffix == ".py":
        return open(notebook).readlines()

    json_notebook = json.load(open(notebook))
    for cell in json_notebook["cells"]:
        if cell["cell_type"] == "code":
            for src_line in cell["source"]:
                ret += [src_line]

    return ret


def extract_previous_stages_from_notebook(notebook: Path):
    if not notebook.exists():
        return []

    ret = []
    if notebook.suffix == ".sh" or notebook.suffix == ".bash":
        ret = open(notebook).readlines()
        ret = [parse_extraction_of_tira_approach_bash(line) for line in ret]
        return [i for i in ret if i]

    for src_line in __read_src_lines(notebook):
        approach = parse_extraction_of_tira_approach(src_line)
        if approach is not None:
            ret += [approach]

    return ret


def extract_previous_stages_from_docker_image(image: str, command: str = None):
    import tempfile

    from tira.local_execution_integration import LocalExecutionIntegration as Client

    if command is None:
        tira = Client()
        command = tira.extract_entrypoint(image)

    notebook = extract_to_be_executed_notebook_from_command_or_none(command)

    if notebook is None:
        return []

    notebook = shlex.split(notebook)[0]

    tira = Client()

    local_file = tempfile.NamedTemporaryFile().name
    if notebook.endswith(".sh") or notebook.endswith(".bash"):
        local_file += ".sh"
    if notebook.endswith(".py"):
        local_file += ".py"
    tira.export_file_from_software(notebook, local_file, image=image)

    return extract_previous_stages_from_notebook(Path(local_file))


def in_tira_sandbox():
    return "TIRA_INPUT_DATASET" in os.environ


def load_ir_datasets():
    try:
        from ir_datasets.datasets.base import Dataset  # noqa: F401
    except Exception:
        return None

    if in_tira_sandbox():
        from tira.ir_datasets_util import static_ir_dataset

        if os.path.isfile(os.path.join(os.environ["TIRA_INPUT_DATASET"], "rerank.jsonl.gz")) or os.path.isfile(
            os.path.join(os.environ["TIRA_INPUT_DATASET"], "rerank.jsonl")
        ):
            import ir_datasets as original_ir_datasets

            register_rerank_data_to_ir_datasets(os.environ["TIRA_INPUT_DATASET"], "dynamic-ds-in-tira")

            return static_ir_dataset(os.environ["TIRA_INPUT_DATASET"], original_ir_datasets.load("dynamic-ds-in-tira"))

        return static_ir_dataset(os.environ["TIRA_INPUT_DATASET"])
    else:
        try:
            from tira.ir_datasets_util import ir_dataset_from_tira_fallback_to_original_ir_datasets

            return ir_dataset_from_tira_fallback_to_original_ir_datasets()
        except Exception:
            return None


ir_datasets = load_ir_datasets()



import unittest

from tira.check_format import GenIrSimulationFormat, check_format

from . import _ERROR, _OK, EMPTY_OUTPUT, GEN_IR_SIM_OUTPUT_VALID, JSONL_OUTPUT_VALID


class TestGenIrSimulationFormat(unittest.TestCase):
    def test_error_message_on_empty_output(self):
        expected = [_ERROR, "No unique *.jsonl file was found, only the files ['.gitkeep'] were available."]
        actual = check_format(EMPTY_OUTPUT, "GenIR-Simulation")
        self.assertEqual(expected, actual)

    def test_error_message_for_other_jsonl_format(self):
        expected = [_ERROR, 'The json line misses the required field "simulation".']
        actual = check_format(JSONL_OUTPUT_VALID, "GenIR-Simulation")
        self.assertEqual(expected, actual)

    def test_invalid_validator_on_empty_output(self):
        expected = [_OK, "The jsonl file has the correct format."]
        actual = check_format(GEN_IR_SIM_OUTPUT_VALID, "GenIR-Simulation")
        self.assertEqual(expected, actual)

    def test_simulation_alone_is_not_valid(self):
        f = GenIrSimulationFormat()
        with self.assertRaises(ValueError):
            f.fail_if_json_line_is_not_valid({"simulation": {}})

    def test_simulation_without_user_turns_is_not_valid(self):
        f = GenIrSimulationFormat()
        with self.assertRaises(ValueError):
            f.fail_if_json_line_is_not_valid({"simulation": {"configuration": {}}})



import json
import logging
import os
import shutil
import subprocess
import sys
import tarfile
import tempfile
import uuid
from copy import deepcopy
from pathlib import Path

import docker
import pandas as pd

from tira.tirex_tracker import tirex_tracker_mounts_or_none


class LocalExecutionIntegration:
    def __init__(self, tira_client=None):
        self.tira_client = tira_client

    def __normalize_command(self, cmd, evaluator):
        to_normalize = {
            "inputRun": "/tira-data/input-run",
            "outputDir": "/tira-data/output",
            "inputDataset": "/tira-data/input",
        }

        if "inputRun" in cmd and evaluator:
            to_normalize["outputDir"] = "/tira-data/eval_output"
            to_normalize["inputDataset"] = "/tira-data/input_truth"

        for k, v in to_normalize.items():
            cmd = cmd.replace("$" + k, v).replace("${" + k + "}", v)

        return cmd

    def __normalize_path(self, path):
        beheaded_path = path if not str(path).startswith("$PWD") else str(path)[4:]
        beheaded_path = (
            os.path.abspath(".")
            if len(beheaded_path) == 0 or beheaded_path in ["/", "."]
            else beheaded_path[0].replace("/", "") + beheaded_path[1:]
        )
        return beheaded_path

    def construct_verbosity_output(self, input_dir, output_dir, image, command, original_args):
        beheaded_input_dir = self.__normalize_path(input_dir)
        beheaded_output_dir = self.__normalize_path(output_dir)
        tira_run_input_dir = beheaded_input_dir if not str(beheaded_input_dir) == str(os.path.abspath(".")) else None
        tira_run_output_dir = (
            beheaded_output_dir if not str(beheaded_output_dir) == str(os.path.abspath("tira-output")) else None
        )
        tira_run_python_args = {k: v for k, v in original_args.items()}
        if type(tira_run_python_args["input_dir"]) is str and tira_run_python_args["input_dir"].startswith("$PWD"):
            tira_run_python_args["input_dir"] = self.__normalize_path(tira_run_python_args["input_dir"])
        if type(tira_run_python_args["output_dir"]) is str and tira_run_python_args["output_dir"].startswith("$PWD"):
            tira_run_python_args["output_dir"] = self.__normalize_path(tira_run_python_args["output_dir"])
        tira_run_args = [(k, v if type(v) is not str else f'"{v}"') for k, v in tira_run_python_args.items()]
        tira_run_args = [
            i.strip() for i in ["" if not tira_run_python_args[k] else f"{k}={v}" for k, v in tira_run_args] if i
        ]

        return {
            "tira-run-cli": "tira-run "
            + ("" if not tira_run_input_dir else f"--input-directory {tira_run_input_dir} ")
            + ("" if not tira_run_output_dir else f"--output-directory {tira_run_output_dir} ")
            + (
                f'--approach {original_args["identifier"]} '
                if "identifier" in original_args and original_args["identifier"] is not None
                else f'--image {image} --command \'{original_args["command"]}\''
            ),
            "tira-run-python": "tira.run(" + ", ".join(tira_run_args) + ")",
            "docker": (
                f"docker run --rm -ti -v {input_dir}:/tira-data/input:ro -v {output_dir}:/tira-data/output:rw"
                f" --entrypoint sh {image} -c '{command}'"
            ),
        }

    def build_docker_image(self, path, tag, dockerfile):
        image_build_code = subprocess.call(["docker", "build", "-f", str(dockerfile), "-t", str(tag), str(path)])

        if image_build_code != 0:
            raise ValueError(
                f"Building the docker image failed with error code {image_build_code}. See above for details."
            )

        print("\n\n Image build successfully.\n\n")

    def ensure_image_available_locally(self, image, client=None):
        try:
            output = subprocess.check_output(["docker", "images", "-q", image])
            if len(output) > 0:
                return
        except Exception:
            pass

        if client:
            try:
                if any(image in str(i) for i in client.images.list()):
                    return
            except Exception:
                pass

        print("# Pull Image\n\n")
        image_pull_code = subprocess.call(["docker", "pull", image])

        if image_pull_code != 0 and "GITHUB_ACTION" in os.environ:
            print("Skip pulling of image because everything is executed within github.")
            return

        if image_pull_code != 0:
            if client is None:
                client = self.__docker_client()
            try:
                client.images.pull(image)
                image_pull_code = 0
            except Exception:
                image_pull_code = 1

        if image_pull_code != 0:
            raise ValueError(
                f"Image could not be successfully pulled. Got return code {image_pull_code}. (expected 0.)"
            )

        print("\n\n Image pulled successfully.\n\nI will now run the software.\n\n")

    def extract_entrypoint(self, image):
        image_name = image
        docker_client = self.__docker_client()
        self.ensure_image_available_locally(image_name, docker_client)
        image = docker_client.images.get(image_name)
        ret = image.attrs["Config"]["Entrypoint"]
        if not ret:
            return None

        for i in deepcopy(ret):
            if i.startswith("[") and i.endswith("]"):
                print(i)
                ret = json.loads(i)
                break

        return self.make_command_absolute(image_name, " ".join(ret))

    def make_command_absolute(self, image_name, command):
        from tira.third_party_integrations import extract_to_be_executed_notebook_from_command_or_none

        executable = extract_to_be_executed_notebook_from_command_or_none(command)

        if not executable or executable.startswith("/") or executable.startswith("'/") or executable.startswith('"/'):
            return command
        else:
            return command.replace(
                executable,
                (self.docker_image_work_dir(image_name) + "/" + executable).replace("//", "/").replace("/./", "/"),
            )

    def __docker_linux_sockets(self):
        ret = [os.path.expanduser("~/.docker/desktop/docker.sock"), "/run/podman/podman.sock"]

        try:
            ret += ["/run/user/" + str(os.getuid()) + "/podman/podman.sock"]
        except Exception:
            pass

        return ret + ["/var/run/docker.sock"]

    def docker_is_installed_failsave(self):
        return self.__docker_client() is not None

    def __docker_client(self):
        try:
            environ = os.environ.copy()
            if sys.platform == "linux" and "DOCKER_HOST" not in environ:
                for docker_socket in self.__docker_linux_sockets():
                    if os.path.exists(docker_socket):
                        environ["DOCKER_HOST"] = "unix://" + docker_socket

                if "DOCKER_HOST" in environ:
                    logging.warn(
                        "Set DOCKER_HOST to '"
                        + environ["DOCKER_HOST"]
                        + "'. Prevent this by explicitly setting the environment variable DOCKER_HOST."
                    )

            client = docker.from_env(environment=environ)

            assert len(client.images.list()) >= 0
            assert len(client.containers.list()) >= 0
            return client
        except Exception as e:
            raise ValueError("It seems like docker is not installed?", e)

    def docker_client_is_authenticated(self, client=None):
        if not client:
            client = self.__docker_client()

        auth_config = client.api._auth_configs
        if (
            not self.tira_client.docker_registry()
            or "auths" not in auth_config
            or self.tira_client.docker_registry() not in auth_config["auths"]
        ):
            return False

        auth_config = auth_config["auths"][self.tira_client.docker_registry()]

        if not auth_config or "username" not in auth_config or "password" not in auth_config:
            return False

        login_response = client.login(
            username=auth_config["username"],
            password=auth_config["password"],
            registry=self.tira_client.docker_registry(),
        )

        return (
            "username" in login_response
            and "password" in login_response
            and auth_config["username"] == login_response["username"]
            and auth_config["password"] == login_response["password"]
        ) or ("Status" in login_response and "login succeeded" == login_response["Status"].lower())

    def login_docker_client(self, task_name, team_name, client=None):
        if not client:
            client = self.__docker_client()

        if self.docker_client_is_authenticated(client):
            return True

        docker_user, docker_password, docker_registry = self.tira_client.docker_credentials(task_name, team_name)

        if not docker_user or not docker_password or not docker_registry:
            print('Please login. Run "tira-cli login --token YOUR-TOKEN-HERE"')
            raise ValueError('Please login. Run "tira-cli login --token YOUR-TOKEN-HERE"')

        # print('Login: ', docker_user, docker_password, docker_registry)
        login_response = client.login(username=docker_user, password=docker_password, registry=docker_registry)

        if "Status" not in login_response or "login succeeded" != login_response["Status"].lower():
            print('Credentials are not valid, please run "tira-cli login --token YOUR-TOKEN-HERE"')
            raise ValueError(f"Login was not successfull, got: {login_response}")

        return True

    def run_and_return_tira_execution(self, task, dataset_directory, team, system_details):
        image = system_details["public_image_name"]
        command = system_details["command"]
        input_docker_software = system_details["input_docker_software"]
        docker_software_id_to_output = {}

        dataset_name = dataset_directory.split("/")[-1]
        log_file = Path(f"{self.tira_client.tira_cache_dir}/.archived/local-executions.jsonl")

        if not log_file.exists():
            with open(log_file, "w") as f:
                f.write("")

        with open(log_file, "r") as f:
            for l in f:
                try:
                    l = json.loads(l)
                    if (
                        l["dataset_directory"] == dataset_directory
                        and l["team"] == team
                        and l["task"] == task
                        and l["image"] == image
                        and l["command"] == command
                    ):
                        return l
                except:
                    pass

        if input_docker_software:
            for i in input_docker_software:
                if i["docker_software_id"] not in docker_software_id_to_output:
                    docker_software_id_to_output[i["docker_software_id"]] = self.run_and_return_tira_execution(
                        task, dataset_directory, team, i
                    )

        run_id = str(uuid.uuid4())
        output_dir = (
            Path(self.tira_client.tira_cache_dir) / "extracted_runs" / task / dataset_name / team / run_id / "output"
        )

        ret = {
            "task": task,
            "dataset": dataset_name,
            "dataset_directory": dataset_directory,
            "team": team,
            "run_id": run_id,
            "image": image,
            "command": command,
            "docker_software_id_to_output": docker_software_id_to_output,
            "output_dir": str(output_dir),
        }

        output_dir.mkdir(parents=True)
        self.run(
            image=image,
            command=command,
            input_dir=dataset_directory,
            output_dir=output_dir,
            input_run=(
                None
                if len(docker_software_id_to_output) != 1
                else list(docker_software_id_to_output.values())[0]["output_dir"]
            ),
            docker_software_id_to_output={k: v["output_dir"] for k, v in docker_software_id_to_output.items()},
        )

        if (output_dir / ".tracking-results.yml").exists():
            shutil.move(output_dir / ".tracking-results.yml", output_dir.parent / ".tracking-results.yml")

        with open(log_file, "a") as f:
            f.write(json.dumps(ret) + "\n")

        return ret

    def tirex_tracker_available_in_docker_image(self, image: str, client: "Optional[DockerClient]" = None) -> bool:
        if client is None:
            client = self.__docker_client()

        volumes = tirex_tracker_mounts_or_none()
        if not volumes:
            return False

        try:
            help_response = client.containers.run(
                image,
                entrypoint="sh",
                command="-c './tracked --help'",
                volumes=volumes,
                detach=False,
                remove=True,
                network_disabled=True,
            ).decode("UTF-8")
        except docker.errors.ContainerError:
            return False

        return "Measures runtime, energy, and many other" in help_response

    def evaluate(self, eval_dir: Path, output_dir: Path, allow_network: bool, evaluate: dict, client=None):
        if not client:
            client = self.__docker_client()

        evaluation_volumes = {str(eval_dir): {"bind": "/tira-data/eval_output", "mode": "rw"}}

        if type(evaluate) is dict and evaluate["evaluator_id"]:
            evaluation_volumes[str(evaluate["truth_directory"])] = {"bind": "/tira-data/input_truth", "mode": "ro"}
            evaluation_volumes[str(output_dir)] = {"bind": "/tira-data/input-run", "mode": "ro"}

            evaluate, image, command = (
                None,
                evaluate["evaluator_git_runner_image"],
                evaluate["evaluator_git_runner_command"],
            )
        elif type(evaluate) is not str:
            raise ValueError("ToDo Implement this case. I.e., set evaluate variable")

        if image is None or command is None:
            evaluate, image, command = self.__extract_image_and_command(evaluate, evaluator=True)

        command = self.__normalize_command(command, True)
        logging.debug(
            f"Evaluate software with: docker run --rm -ti -v {output_dir}:/tira-data/input -v"
            f" {output_dir}/:/tira-data/output --entrypoint sh {image} -c '{command}'"
        )

        container = client.containers.run(
            image,
            entrypoint="sh",
            command=f'-c "{command}; sleep .1"',
            volumes=evaluation_volumes,
            detach=True,
            remove=True,
            network_disabled=not allow_network,
        )

        for line in container.attach(stdout=True, stream=True, logs=True):
            print(line.decode("utf-8"), flush=True)

    def run(
        self,
        identifier=None,
        image=None,
        command=None,
        input_dir=None,
        output_dir=None,
        evaluate=False,
        dry_run=False,
        docker_software_id_to_output=None,
        software_id=None,
        allow_network=False,
        input_run=None,
        additional_volumes=None,
        eval_dir="tira-evaluation",
        gpu_count=0,
    ):
        previous_stages = []
        original_args = {
            "identifier": identifier,
            "image": image,
            "command": command,
            "input_dir": input_dir,
            "output_dir": output_dir,
            "evaluate": evaluate,
            "dry_run": dry_run,
            "docker_software_id_to_output": docker_software_id_to_output,
            "software_id": software_id,
        }
        s_id = "unknown-software-id"
        if image is None or command is None:
            ds = self.tira_client.docker_software(approach=identifier, software_id=software_id)
            image, command, s_id, previous_stages = (
                ds["tira_image_name"],
                ds["command"],
                ds["id"],
                ds["ids_of_previous_stages"],
            )
        if not dry_run:
            client = self.__docker_client()

        command = self.__normalize_command(command, False)

        if not input_dir or not output_dir:
            raise ValueError("please pass input_dir and output_dir")

        input_dir = os.path.abspath(input_dir) if not str(input_dir).startswith("$PWD") else input_dir
        output_dir = os.path.abspath(output_dir) if not str(output_dir).startswith("$PWD") else output_dir

        docker_software_id_to_output = (
            {} if not docker_software_id_to_output else deepcopy(docker_software_id_to_output)
        )

        for previous_stage in previous_stages:
            if previous_stage in docker_software_id_to_output.keys():
                continue

            tmp_prev_stages = self.run(
                software_id=previous_stage,
                identifier=None,
                image=None,
                command=None,
                input_dir=input_dir,
                evaluate=False,
                dry_run=dry_run,
                output_dir=tempfile.TemporaryDirectory("-staged-execution-" + previous_stage).name + "/output",
                docker_software_id_to_output=docker_software_id_to_output,
            )
            for k, v in tmp_prev_stages.items():
                docker_software_id_to_output[k] = v

        verbose_data = self.construct_verbosity_output(input_dir, output_dir, image, command, original_args)
        logging.debug(
            f'Docker:\n\t{verbose_data["docker"]}\n\ntira-run'
            f' (python):\n\t{verbose_data["tira-run-python"]}\n\ntira-run (CLI):\n\t{verbose_data["tira-run-cli"]}\n\n'
        )

        if dry_run:
            return verbose_data

        volumes = {
            str(input_dir): {"bind": "/tira-data/input", "mode": "ro"},
            str(output_dir): {"bind": "/tira-data/output", "mode": "rw"},
        }

        if input_run:
            volumes[str(input_run)] = {"bind": "/tira-data/input-run", "mode": "ro"}

        for k, v in docker_software_id_to_output.items():
            volumes[str(os.path.abspath(v))] = {"bind": "/tira-data/input-run", "mode": "ro"}

        if additional_volumes:
            for v in additional_volumes:
                volume_dir, volume_bind, volume_mode = v.split(":")
                volume_dir = str(os.path.abspath(volume_dir))
                if volume_dir in volumes:
                    raise ValueError(f"Volume to mount is multiple times defined: {volume_dir}")
                volumes[volume_dir] = {"bind": volume_bind, "mode": volume_mode}

        self.ensure_image_available_locally(image, client)
        environment = {
            "outputDir": "/tira-data/output",
            "inputDataset": "/tira-data/input",
            "TIRA_DATASET_ID": "id",
            "TIRA_OUTPUT_DIR": "/tira-data/output",
            "TIRA_INPUT_DATASET": "/tira-data/input",
        }

        if input_run:
            environment["inputRun"] = "/tira-data/input-run"

        device_requests = []
        if gpu_count != 0:
            device_requests = [docker.types.DeviceRequest(count=gpu_count, capabilities=[["gpu"]])]

        entrypoint = "sh"
        entrypoint_flags = "-c"
        if self.tirex_tracker_available_in_docker_image(image, client):
            volumes.update(tirex_tracker_mounts_or_none())
            entrypoint = "/tracked"
            entrypoint_flags = "-o /tira-data/output/.tracking-results.yml -f irmetadata"

        container = client.containers.run(
            image,
            entrypoint=entrypoint,
            command=f'{entrypoint_flags} "{command}; sleep .1"',
            environment=environment,
            volumes=volumes,
            detach=True,
            remove=True,
            network_disabled=not allow_network,
            device_requests=device_requests,
        )

        for line in container.attach(stdout=True, stream=True, logs=True):
            print(line.decode("utf-8"))

        if evaluate:
            self.evaluate(eval_dir, output_dir, allow_network, client)

        if evaluate:
            approach_name = identifier if identifier else f'"{command}"@{image}'
            eval_results = {"approach": approach_name, "evaluate": evaluate}
            eval_results.update(self.load_output_of_directory(Path(eval_dir), evaluation=True))
            return self.load_output_of_directory(Path(eval_dir)), pd.DataFrame([eval_results])
        else:
            docker_software_id_to_output[s_id] = output_dir
            return docker_software_id_to_output

    def docker_image_work_dir(self, image):
        image = self.__docker_client().images.get(image).attrs["Config"]
        return "/" + image.get("WorkingDir", "")

    def export_file_from_software(self, container_path, host_path, identifier=None, image=None, software_id=None):
        """
        Export a file specified by container_path' from a software to the host_path at the host.
        """
        if image is None:
            ds = self.tira_client.docker_software(approach=identifier, software_id=software_id)
            image = ds["tira_image_name"]

        client = self.__docker_client()
        self.ensure_image_available_locally(image, client)
        docker_container = client.containers.create(image)
        strm, stat = docker_container.get_archive(container_path, None)

        with open(host_path, "wb") as f:
            for i in strm:
                f.write(i)

        tf = tarfile.open(host_path, mode="r")
        tf = tf.extractfile(stat["name"]).read()

        with open(host_path, "wb") as f:
            f.write(tf)

        docker_container.remove()

    def export_submission_from_jupyter_notebook(self, notebook):
        if not os.path.isfile(notebook):
            print(f"The notebook {notebook} does not exist. I can not continue.", file=sys.stderr)

            if "/" in notebook:
                try:
                    notebook = "/".join(notebook.split("/")[:-1])
                    print(
                        f"The directory {notebook} contains the files {os.listdir(notebook)}. Maybe you did mean one of"
                        " those?",
                        file=sys.stderr,
                    )
                except Exception:
                    pass

            return None

        ret = []

        ret += [
            "TIRA_COMMAND=/workspace/run-pyterrier-notebook.py --input ${TIRA_INPUT_DIRECTORY} --output"
            " ${TIRA_OUTPUT_DIRECTORY} --notebook /workspace/" + notebook.split("/")[-1]
        ]
        notebook_content = json.load(open(notebook, "r"))
        print(f"TODO: process content {notebook_content}.")

        return "\n".join(ret)

    def normalize_image_name(self, image, required_prefix):
        if required_prefix and not image.startswith(required_prefix):
            ret = (
                (required_prefix + "/" + (image + ":").split(":")[0][:10]).replace("//", "/")
                + ":"
                + (image.split(":")[-1] if ":" in image else "latest")
            )
            return ret.replace("-:", ":").replace("::", ":").replace("/:", ":")
        else:
            return image

    def show_docker_progress(self, line, tasks):
        from tqdm import tqdm

        if "status" not in line or (
            line["status"] != "Downloading" and line["status"] != "Extracting" and line["status"] != "Pushing"
        ):
            # skip other statuses
            return

        id = f'{line["id"]}--{line["status"]}'

        if id not in tasks:
            tasks[id] = tqdm(position=len(tasks) + 1, desc=f'{line["status"]} ({line["id"]}):')

        try:
            tasks[id].update(line["progressDetail"]["total"])
        except Exception:
            pass

    def push_image(self, image, required_prefix=None, task_name=None, team_name=None):
        client = self.__docker_client()
        if not self.docker_client_is_authenticated(client):
            self.login_docker_client(task_name, team_name, client)
        new_image = image
        if required_prefix and not image.startswith(required_prefix):
            new_image = self.normalize_image_name(image, required_prefix)
            print(f'I tag the image "{image}" as "{new_image}" for upload to TIRA (only internal).')
            client.images.get(image).tag(new_image)
            image = new_image

        tasks = {}
        push_response = ""
        for line in client.images.push(image, stream=True, decode=True):
            push_response += str(line)
            self.show_docker_progress(line, tasks)

        if "error" in push_response:
            print(push_response)
            raise ValueError("Could not push image")
        return new_image



import unittest

from tira.local_execution_integration import LocalExecutionIntegration
from tira.tirex_tracker import find_tirex_tracker_executable_or_none


class TirexTrackerTest(unittest.TestCase):
    def test_path_to_tirex_tracker_exists(self):
        actual = find_tirex_tracker_executable_or_none()
        self.assertIsNotNone(actual)
        self.assertTrue(actual.exists())

    def test_tirex_tracker_not_available_in_bash_image(self):
        tira = LocalExecutionIntegration()
        actual = tira.tirex_tracker_available_in_docker_image("bash:alpine3.16")
        self.assertFalse(actual)

    def test_tirex_tracker_is_available_in_debian_image(self):
        tira = LocalExecutionIntegration()
        actual = tira.tirex_tracker_available_in_docker_image("debian:trixie-slim")
        self.assertTrue(actual)
